# TODO-1: maybe ref all to more simple&
# TODO: ref all signals into one with [int]

# =====================================================================================================================
from typing import *
import time
from threading import Thread

from PyQt5.QtCore import *
from PyQt5.QtWidgets import *

from base_aux.base_values.m4_primitives import LAMBDA_TRUE
from base_aux.aux_datetime.m1_sleep import Sleep

from base_aux.pyqt.m0_static import COLOR_TUPLE_RGB

from base_aux.base_types.m2_info import ObjectInfo


# SET ============================================================================================================
class Dialogs:
    """
    GOAL
    ----
    attempt to keep all available dialogs for current project in one place!
    so to be sure there are no any other available/not defined!

    use only one active wgt for same purpose!
    """
    WGTS: dict[str, QMessageBox] = {}

    @classmethod
    def _apply_new__or_activate_last(cls, name: str, wgt: QMessageBox) -> bool:
        if name in cls.WGTS and cls.WGTS[name] and cls.WGTS[name].isVisible():
            return False

        # try:
        #     # cls.WGTS[name].close()      # not working!
        #     # cls.WGTS[name].destroy()    # not working!
        #     # cls.WGTS[name].done(1)      # not working!
        #     # cls.WGTS[name].done(0)      # not working!
        #     # cls.WGTS[name].accept()      # not working!
        #     cls.WGTS[name].activateWindow()      # not working!
        #     # cls.WGTS[name].reject()      # not working!
        #     return False
        # except:
        #     pass

        cls.WGTS[name] = wgt
        return True

    @classmethod
    def info__about(cls, *args) -> int:
        # 1way=SIMPLE=direct meth on class ----------------
        # answer = QMessageBox.information(
        #     None,
        #     "About Program",
        #     (
        #         "LLC CompanyName,\n"
        #         "Program name\n"
        #         "(purpose)"
        #      )
        # )

        # 2way=WgtObject -------------------------
        wgt = QMessageBox()
        ObjectInfo(wgt).print()
        if not cls._apply_new__or_activate_last(name="info__about", wgt=wgt):
            return 1024
        wgt.setMaximumWidth(1000)
        answer = wgt.information(
            None,
            "About Program",
            (
                "LLC CompanyName,\n"
                "Program name\n"
                "(purpose)"
             )
        )

        ObjectInfo(wgt).print()
        # return always 1024
        return answer


# SIMPLEST ============================================================================================================
def try__info():
    app = QApplication([])
    print(Dialogs.info__about())
    print(Dialogs.info__about())
    app.exec()


# OTHERS ==============================================================================================================
class MessageDialog(QDialog):   # Шихалиев???
    """
    just found in corporate repo from not a really programming department
    """
    def __init__(self, text, parent=None):
        super(MessageDialog, self).__init__(parent)

        layout = QVBoxLayout(self)
        layout.addWidget(QLabel(text))

        # OK and Cancel buttons
        self.buttons = QDialogButtonBox(
            QDialogButtonBox.Ok | QDialogButtonBox.Cancel,
            Qt.Horizontal, self)
        layout.addWidget(self.buttons)

        self.buttons.accepted.connect(self.accept)
        self.buttons.rejected.connect(self.reject)

    # static method to create the dialog and return (date, time, accepted)
    @staticmethod
    def info(text, parent=None):
        dialog = MessageDialog(text, parent)
        result = dialog.exec_()
        return (result == QDialog.Accepted)


# =====================================================================================================================
class GuiDialog(QWidget):
    """
    allow to use Gui Interactive from independent console!

    ATTENTION!!!
        YOU MUST INSTANTIATE OBJECT IN MAIN THREAD!!!
        SO IT WILL BE WORKING IN ANY THREAD!!! otherwise will not!
        dont need to save it! only start!

    IMPORTANT:
        1. ALL methods if ask accept/reject must return TRUE if clicked OK/Apply buttons! otherwise False!
        2. first parameter in any

    EXAMPLE usage:
        GuiDialog().simple_1_info()



    FOR AUTOACCEPTION    !!!!!!!!!!!!!!!IN CASE OF FREEZE!!!!!!!!!!!!!!!!!!!
        Only one reason is - incondition state in result of autoaccept_link
        for example
            WRONG -     [GuiDialog().simple_1_info(message, autoaccept_link=lambda: self.fu_get_current_v12())]
            CORRECT -   [GuiDialog().simple_1_info(message, autoaccept_link=lambda: bool(self.fu_get_current_v12()))]
    """
    s_type0 = pyqtSignal(str, str)  # (msg, title)
    s_type1 = pyqtSignal(str, str)
    s_type2 = pyqtSignal(str, str)
    s_type3 = pyqtSignal(str, str)
    s_type4 = pyqtSignal(str, str)
    s_type5 = pyqtSignal(str, str)
    s_type6 = pyqtSignal(str, str)
    s_type7 = pyqtSignal(str, str)

    WGT_ROOT = None  # if contain object - will be used signal!
    WGT_ROOT_NAME = "WGT_ROOT_NAME"

    WGT_NOLOCK_NOANSWER = None

    # def __new__(cls, *args, **kwargs):
    #     if cls.WGT_ROOT is None:
    #         cls.WGT_ROOT = super().__new__(cls, *args, **kwargs)
    #     return super().__new__(cls, *args, **kwargs)

    ANSWER_READY = None
    ANSWER_LAST = None
    ANSWER_LAST_AUTOACCEPTED = None

    def __init__(self, _create_singleton_obj=None):
        """
        :param _create_singleton_obj: DONT USE IT!!! it must start only auto in deep code by self!
        """
        self.app = None

        if _create_singleton_obj:
            self._init_qapplication_or_root_dialog_obj()
            super().__init__()
            self.setObjectName(self.WGT_ROOT_NAME)

            self.s_type0.connect(self._only_simple_0_nolock_noanswer)
            self.s_type1.connect(self._only_simple_1_info)
            self.s_type2.connect(self._only_simple_2_question)
            self.s_type3.connect(self._only_simple_3_critical)
            self.s_type4.connect(self._only_input_string)
            self.s_type5.connect(self._only_select_file)
            self.s_type6.connect(self._only_select_font)
            self.s_type7.connect(self._only_select_color)

        elif __class__.WGT_ROOT is None:
            __class__.WGT_ROOT = __class__(_create_singleton_obj=True)

    # CLOSE/ACCEPT ====================================================================================================
    @classmethod
    def close_wgt_nolock_noanswer(cls):
        cls.WGT_NOLOCK_NOANSWER = None

    @classmethod
    def accept_all_dialogs(cls):
        while True:  # need to accept all Dialogs without overlapping!
            top_level_widgets = QApplication.topLevelWidgets()
            # print(f"{top_level_widgets=}")
            if len(top_level_widgets) > 0:
                for wgt_i in top_level_widgets:
                    # print(f"{wgt_i.objectName()=}/{wgt_i=}")
                    if hasattr(wgt_i,
                               "accept") and wgt_i.parent() and wgt_i.parent().objectName() == cls.WGT_ROOT_NAME:
                        # print(f"ACCEPT-START={wgt_i.parent().objectName()=}/{wgt_i.objectName()=}/{wgt_i=}/{wgt_i.parent()=}")
                        try:
                            time.sleep(0.3)  # without sleep it will work instable!!! ones a 10 steps will get freez!
                            wgt_i.accept()
                        except:
                            pass

                        return
                    # if wgt_i.objectName() == self.WGT_ROOT_NAME:
                    #     print(f"ЗАКРЫТИЕ!!! - СТАРТ")
                    #     wgt_i.close()
                    #     print(f"ЗАКРЫТИЕ!!! - ФИНИШ!!!")
            break
        return True

    @classmethod
    def autoaccept_start_thread(cls, autoaccept_func_link):
        thread = Thread(target=lambda: cls.autoaccept_start(autoaccept_func_link), daemon=True)
        thread.start()

    @classmethod
    def autoaccept_start(cls, autoaccept_func_link):
        """close all dialogs in all threads!!!"""
        result = False
        while not cls.ANSWER_READY:
            try:
                result = autoaccept_func_link()
            except:
                pass

            if result:
                cls.accept_all_dialogs()
                cls.ANSWER_LAST_AUTOACCEPTED = result
                return True
            else:
                time.sleep(1)

    # =================================================================================================================
    def _init_qapplication_or_root_dialog_obj(self):
        """
        init QApplication only if was not called! otherwise will incorrect working at sequence starting!!!
        :return:

        CHECK:
            enaugh only on one Thread project!!!
                QApplication.applicationName() - on child thread will show not blank!!!
                QApplication.hasPendingEvents() - on child thread will show False!!!


            # print("111111111111\n"*10)
            # UFU.obj_show_attr_all(QApplication, miss_names=["aboutQt", "beep", "fontMetrics", "pyqtConfigure", "exec", "exec_"])
            ////////////////////////////////////////////////// obj_show_attr_all //////////////////////////////////////////////////
            ////////// _parents_list=[] nested_level=[0] source=<class 'PyQt5.QtWidgets.QApplication'> //////////
            meth=[.ColorSpec]                       value=[0]
            attr=[.CustomColor]                     value=[1]
            attr=[.ManyColor]                       value=[2]
            attr=[.NormalColor]                     value=[0]
            meth=[.__class__]                       value=[<class 'sip.wrappertype'>]
            attr=[.__delattr__]                     value=[***MISSED DANGER***]
            obj=[.__dict__]                         value=[{'__module__': 'PyQt5.QtWidgets', '__doc__': 'QApplication(List[str])', 'ColorSpec': <class 'PyQt5.QtWidgets.QApplication.ColorSpec'>, 'aboutQt': <built-in method aboutQt>, 'activeModalWidget': <built-in method activeModalWidget>, 'activePopupWidget': <built-in method activePopupWidget>, 'activeWindow': <built-in method activeWindow>, 'alert': <built-in method alert>, 'allWidgets': <built-in method allWidgets>, 'autoSipEnabled': <built-in method autoSipEnabled>, 'beep': <built-in method beep>, 'childEvent': <built-in method childEvent>, 'closeAllWindows': <built-in method closeAllWindows>, 'colorSpec': <built-in method colorSpec>, 'connectNotify': <built-in method connectNotify>, 'cursorFlashTime': <built-in method cursorFlashTime>, 'customEvent': <built-in method customEvent>, 'desktop': <built-in method desktop>, 'disconnectNotify': <built-in method disconnectNotify>, 'doubleClickInterval': <built-in method doubleClickInterval>, 'event': <built-in method event>, 'exec': <built-in method exec>, 'exec_': <built-in method exec_>, 'focusWidget': <built-in method focusWidget>, 'font': <built-in method font>, 'fontMetrics': <built-in method fontMetrics>, 'globalStrut': <built-in method globalStrut>, 'isEffectEnabled': <built-in method isEffectEnabled>, 'isSignalConnected': <built-in method isSignalConnected>, 'keyboardInputInterval': <built-in method keyboardInputInterval>, 'notify': <built-in method notify>, 'palette': <built-in method palette>, 'receivers': <built-in method receivers>, 'sender': <built-in method sender>, 'senderSignalIndex': <built-in method senderSignalIndex>, 'setActiveWindow': <built-in method setActiveWindow>, 'setAutoSipEnabled': <built-in method setAutoSipEnabled>, 'setColorSpec': <built-in method setColorSpec>, 'setCursorFlashTime': <built-in method setCursorFlashTime>, 'setDoubleClickInterval': <built-in method setDoubleClickInterval>, 'setEffectEnabled': <built-in method setEffectEnabled>, 'setFont': <built-in method setFont>, 'setGlobalStrut': <built-in method setGlobalStrut>, 'setKeyboardInputInterval': <built-in method setKeyboardInputInterval>, 'setPalette': <built-in method setPalette>, 'setStartDragDistance': <built-in method setStartDragDistance>, 'setStartDragTime': <built-in method setStartDragTime>, 'setStyle': <built-in method setStyle>, 'setStyleSheet': <built-in method setStyleSheet>, 'setWheelScrollLines': <built-in method setWheelScrollLines>, 'setWindowIcon': <built-in method setWindowIcon>, 'startDragDistance': <built-in method startDragDistance>, 'startDragTime': <built-in method startDragTime>, 'style': <built-in method style>, 'styleSheet': <built-in method styleSheet>, 'timerEvent': <built-in method timerEvent>, 'topLevelAt': <built-in method topLevelAt>, 'topLevelWidgets': <built-in method topLevelWidgets>, 'wheelScrollLines': <built-in method wheelScrollLines>, 'widgetAt': <built-in method widgetAt>, 'windowIcon': <built-in method windowIcon>, 'CustomColor': 1, 'ManyColor': 2, 'NormalColor': 0, 'focusChanged': <unbound PYQT_SIGNAL focusChanged(QWidget*,QWidget*)>}]
            meth=[.__dir__]                         value=[<method '__dir__' of 'object' base_types>]
            attr=[.__doc__]                         value=[QApplication(List[str])]
            attr=[.__enter__]                       value=[***MISSED DANGER***]
            meth=[.__eq__]                          value=[<slot wrapper '__eq__' of 'object' base_types>]
            attr=[.__exit__]                        value=[***MISSED DANGER***]
            meth=[.__format__]                      value=[<method '__format__' of 'object' base_types>]
            meth=[.__ge__]                          value=[<slot wrapper '__ge__' of 'object' base_types>]
            meth=[.__getattr__]                     value=[<built-in function __getattr__>]
            meth=[.__getattribute__]                value=[<slot wrapper '__getattribute__' of 'object' base_types>]
            meth=[.__gt__]                          value=[<slot wrapper '__gt__' of 'object' base_types>]
            meth=[.__hash__]                        value=[<slot wrapper '__hash__' of 'object' base_types>]
            attr=[.__init__]                        value=[***MISSED DANGER***]
            attr=[.__init_subclass__]               value=[***MISSED DANGER***]
            meth=[.__le__]                          value=[<slot wrapper '__le__' of 'object' base_types>]
            meth=[.__lt__]                          value=[<slot wrapper '__lt__' of 'object' base_types>]
            attr=[.__module__]                      value=[PyQt5.QtWidgets]
            meth=[.__ne__]                          value=[<slot wrapper '__ne__' of 'object' base_types>]
            attr=[.__new__]                         value=[***MISSED DANGER***]
            attr=[.__reduce__]                      value=[***MISSED DANGER***]
            attr=[.__reduce_ex__]                   value=[***MISSED DANGER***]
            meth=[.__repr__]                        value=[<slot wrapper '__repr__' of 'object' base_types>]
            attr=[.__setattr__]                     value=[***MISSED DANGER***]
            meth=[.__sizeof__]                      value=[<method '__sizeof__' of 'object' base_types>]
            meth=[.__str__]                         value=[<slot wrapper '__str__' of 'object' base_types>]
            meth=[.__subclasshook__]                value=[NotImplemented]
            obj=[.__weakref__]                      value=[<attribute '__weakref__' of 'QObject' base_types>]
            attr=[.aboutQt]                         value=[***MISSED SPECIAL***]
            meth=[.aboutToQuit]                     value=[<unbound PYQT_SIGNAL aboutToQuit()>]
            meth=[.activeModalWidget]               value=[None]
            meth=[.activePopupWidget]               value=[None]
            meth=[.activeWindow]                    value=[<gui.main_widget.TestSystemWidget object at 0x000002880E8FAA60>]
            meth=[.addLibraryPath]                  value=[<built-in function addLibraryPath>]
            meth=[.alert]                           value=[<built-in function alert>]
            meth=[.allWidgets]                      value=[[<PyQt5.QtWidgets.QScrollBar object at 0x0000028811F158B0>, <PyQt5.QtWidgets.QLabel object at 0x0000028810BB5DC0>, <PyQt5.QtWidgets.QWidget object at 0x0000028811F15820>, <PyQt5.QtWidgets.QScrollBar object at 0x0000028811F15670>, <PyQt5.QtWidgets.QWidget object at 0x0000028811F15700>, <PyQt5.QtWidgets.QWidget object at 0x0000028811F159D0>, <PyQt5.QtWidgets.QHeaderView object at 0x0000028811F15A60>, <PyQt5.QtWidgets.QWidget object at 0x0000028811F15790>, <PyQt5.QtWidgets.QWidget object at 0x0000028811F155E0>, <PyQt5.QtWidgets.QSplitter object at 0x000002880EA21550>, <PyQt5.QtWidgets.QWidget object at 0x0000028811F15940>, <gui.authorization_widget.WidgetServerAuthorization object at 0x000002881306CF70>, <PyQt5.QtWidgets.QLabel object at 0x0000028810BBBEE0>, <PyQt5.QtWidgets.QPushButton object at 0x000002880EA00EE0>, <PyQt5.QtWidgets.QPushButton object at 0x000002880EA00790>, <PyQt5.QtWidgets.QFrame object at 0x0000028810BBBB80>, <PyQt5.QtWidgets.QFrame object at 0x0000028811F15550>, <PyQt5.QtWidgets.QPushButton object at 0x000002880EA21700>, <PyQt5.QtWidgets.QPushButton object at 0x000002880EA00A60>, <PyQt5.QtWidgets.QHeaderView object at 0x0000028811F15AF0>, <PyQt5.QtWidgets.QHeaderView object at 0x0000028811F15B80>, <PyQt5.QtWidgets.QWidget object at 0x0000028811F15C10>, <PyQt5.QtWidgets.QScrollBar object at 0x0000028811F15CA0>, <PyQt5.QtWidgets.QLabel object at 0x0000028811F15D30>, <PyQt5.QtWidgets.QFrame object at 0x0000028810BB83A0>, <PyQt5.QtWidgets.QLabel object at 0x0000028810BB80D0>, <gui.widget_progress_bar.WidgetProgressBar object at 0x0000028810BBBDC0>, <PyQt5.QtWidgets.QFrame object at 0x0000028811F15DC0>, <PyQt5.QtWidgets.QWidget object at 0x0000028811F15E50>, <PyQt5.QtWidgets.QLabel object at 0x0000028810BB5B80>, <PyQt5.QtWidgets.QScrollBar object at 0x0000028811F15EE0>, <PyQt5.QtWidgets.QCheckBox object at 0x0000028810BBBAF0>, <PyQt5.QtWidgets.QWidget object at 0x0000028810BBB040>, <PyQt5.QtWidgets.QSplitterHandle object at 0x0000028811F15F70>, <PyQt5.QtWidgets.QPushButton object at 0x0000028810BBB9D0>, <PyQt5.QtWidgets.QPushButton object at 0x000002880EA21940>, <PyQt5.QtWidgets.QSplitterHandle object at 0x0000028811D57040>, <PyQt5.QtWidgets.QPushButton object at 0x0000028810BB5790>, <PyQt5.QtWidgets.QLabel object at 0x0000028810BC1040>, <PyQt5.QtWidgets.QSplitter object at 0x000002880E8FAB80>, <PyQt5.QtWidgets.QLineEdit object at 0x0000028811D570D0>, <PyQt5.QtWidgets.QSplitterHandle object at 0x0000028811D57160>, <PyQt5.QtWidgets.QLabel object at 0x0000028810BB5940>, <PyQt5.QtWidgets.QListView object at 0x0000028811D571F0>, <PyQt5.QtWidgets.QLabel object at 0x0000028810BB8310>, <PyQt5.QtWidgets.QPushButton object at 0x0000028810BBBD30>, <PyQt5.QtWidgets.QWidget object at 0x0000028811D57280>, <PyQt5.QtWidgets.QPushButton object at 0x0000028810BBB4C0>, <PyQt5.QtWidgets.QWidget object at 0x0000028811D57310>, <PyQt5.QtWidgets.QPushButton object at 0x0000028810BB54C0>, <PyQt5.QtWidgets.QWidget object at 0x0000028811D573A0>, <PyQt5.QtWidgets.QPushButton object at 0x0000028813089550>, <PyQt5.QtWidgets.QDesktopWidget object at 0x0000028811D57430>, <PyQt5.QtWidgets.QPushButton object at 0x0000028810BB5310>, <PyQt5.QtWidgets.QLabel object at 0x0000028810BB8A60>, <PyQt5.QtWidgets.QScrollBar object at 0x0000028811D574C0>, <PyQt5.QtWidgets.QLabel object at 0x0000028810BB5A60>, <PyQt5.QtWidgets.QPushButton object at 0x0000028810BBB550>, <PyQt5.QtWidgets.QPushButton object at 0x000002880EA21D30>, <PyQt5.QtWidgets.QLabel object at 0x0000028810BB5D30>, <PyQt5.QtWidgets.QPushButton object at 0x0000028810BBB820>, <PyQt5.QtWidgets.QProgressBar object at 0x0000028810BBB1F0>, <PyQt5.QtWidgets.QWidget object at 0x0000028811D57550>, <PyQt5.QtWidgets.QScrollBar object at 0x0000028811D575E0>, <PyQt5.QtWidgets.QLabel object at 0x0000028810BB5E50>, <PyQt5.QtWidgets.QWidget object at 0x0000028811D57670>, <PyQt5.QtWidgets.QAbstractButton object at 0x0000028811D57700>, <PyQt5.QtWidgets.QWidget object at 0x0000028811D57790>, <PyQt5.QtWidgets.QCheckBox object at 0x0000028810BB8CA0>, <PyQt5.QtWidgets.QComboBox object at 0x0000028810BC1280>, <PyQt5.QtWidgets.QLabel object at 0x000002880EA215E0>, <PyQt5.QtWidgets.QPushButton object at 0x0000028810BBBA60>, <PyQt5.QtWidgets.QScrollBar object at 0x0000028811D57820>, <PyQt5.QtWidgets.QScrollBar object at 0x0000028811D578B0>, <PyQt5.QtWidgets.QLabel object at 0x0000028810BB5820>, <PyQt5.QtWidgets.QLabel object at 0x0000028810BB8160>, <PyQt5.QtWidgets.QLabel object at 0x0000028810BB5550>, <PyQt5.QtWidgets.QWidget object at 0x0000028811D57940>, <PyQt5.QtWidgets.QLabel object at 0x0000028813089310>, <gui.stand_settings_edit_widget.WidgetStandSettings object at 0x000002880EA21040>, <PyQt5.QtWidgets.QSplitterHandle object at 0x0000028811D579D0>, <PyQt5.QtWidgets.QWidget object at 0x0000028811D57A60>, <PyQt5.QtWidgets.QScrollBar object at 0x0000028811D57AF0>, <PyQt5.QtWidgets.QLabel object at 0x0000028813089040>, <PyQt5.QtWidgets.QComboBox object at 0x0000028810BB5670>, <PyQt5.QtWidgets.QSplitterHandle object at 0x0000028811D57B80>, <PyQt5.QtWidgets.QWidget object at 0x0000028811D57C10>, <PyQt5.QtWidgets.QWidget object at 0x0000028811D57CA0>, <PyQt5.QtWidgets.QFrame object at 0x0000028811D57D30>, <PyQt5.QtWidgets.QWidget object at 0x0000028811D57DC0>, <gui.widget_stand.WgtStand object at 0x000002880EA21DC0>, <PyQt5.QtWidgets.QLabel object at 0x0000028810BBBF70>, <PyQt5.QtWidgets.QComboBox object at 0x0000028810BB5280>, <PyQt5.QtWidgets.QSplitterHandle object at 0x0000028811D57E50>, <PyQt5.QtWidgets.QScrollBar object at 0x0000028811D57EE0>, <gui.main_widget.TestSystemWidget object at 0x000002880E8FAA60>, <PyQt5.QtWidgets.QWidget object at 0x0000028811D57F70>, <PyQt5.QtWidgets.QScrollBar object at 0x0000028812D8D040>, <PyQt5.QtWidgets.QLabel object at 0x0000028810BB8040>, <PyQt5.QtWidgets.QLabel object at 0x000002880EA218B0>, <PyQt5.QtWidgets.QTableView object at 0x0000028810BBB790>, <PyQt5.QtWidgets.QScrollBar object at 0x0000028812D8D0D0>, <PyQt5.QtWidgets.QScrollBar object at 0x0000028812D8D160>, <PyQt5.QtWidgets.QScrollBar object at 0x0000028812D8D1F0>, <PyQt5.QtWidgets.QScrollBar object at 0x0000028812D8D280>, <PyQt5.QtWidgets.QScrollBar object at 0x0000028812D8D310>, <PyQt5.QtWidgets.QWidget object at 0x0000028812D8D3A0>, <PyQt5.QtWidgets.QLabel object at 0x000002880EA00DC0>, <PyQt5.QtWidgets.QWidget object at 0x0000028812D8D430>, <PyQt5.QtWidgets.QComboBox object at 0x0000028810BB5430>, <gui.file_widget.WgtUser_Profile object at 0x0000028810BB85E0>, <PyQt5.QtWidgets.QLabel object at 0x0000028810BB8AF0>, <PyQt5.QtWidgets.QWidget object at 0x0000028812D8D4C0>, <PyQt5.QtWidgets.QLabel object at 0x0000028810BB5AF0>, <PyQt5.QtWidgets.QPushButton object at 0x000002880EA21790>, <PyQt5.QtWidgets.QWidget object at 0x0000028812D8D550>, <PyQt5.QtWidgets.QPushButton object at 0x000002880EA00820>, <PyQt5.QtWidgets.QWidget object at 0x0000028812D8D5E0>, <PyQt5.QtWidgets.QScrollBar object at 0x0000028812D8D670>, <PyQt5.QtWidgets.QSplitter object at 0x000002887B44D550>, <PyQt5.QtWidgets.QPushButton object at 0x000002880EA00940>, <PyQt5.QtWidgets.QLabel object at 0x0000028810BB51F0>, <PyQt5.QtWidgets.QWidget object at 0x0000028812D8D700>, <PyQt5.QtWidgets.QLabel object at 0x0000028810BB53A0>, <PyQt5.QtWidgets.QScrollBar object at 0x0000028812D8D790>, <PyQt5.QtWidgets.QPushButton object at 0x000002880EA008B0>, <PyQt5.QtWidgets.QFrame object at 0x0000028812D8D820>, <PyQt5.QtWidgets.QWidget object at 0x0000028812D8D8B0>, <PyQt5.QtWidgets.QFrame object at 0x0000028810BBB280>, <PyQt5.QtWidgets.QLineEdit object at 0x00000288130891F0>, <PyQt5.QtWidgets.QLabel object at 0x0000028810BB5C10>, <PyQt5.QtWidgets.QFrame object at 0x0000028810BBB940>, <PyQt5.QtWidgets.QWidget object at 0x0000028812D8D940>, <PyQt5.QtWidgets.QPushButton object at 0x0000028810BB89D0>, <PyQt5.QtWidgets.QLabel object at 0x0000028810BB8EE0>, <PyQt5.QtWidgets.QFrame object at 0x000002880EA00B80>, <PyQt5.QtWidgets.QLabel object at 0x0000028810BB8B80>, <PyQt5.QtWidgets.QPushButton object at 0x000002880EA219D0>, <PyQt5.QtWidgets.QScrollBar object at 0x0000028812D8D9D0>, <PyQt5.QtWidgets.QWidget object at 0x0000028812D8DA60>, <PyQt5.QtWidgets.QPlainTextEdit object at 0x0000028810BB8F70>, <PyQt5.QtWidgets.QLabel object at 0x0000028810BB59D0>, <PyQt5.QtWidgets.QLabel object at 0x0000028810BB5CA0>, <PyQt5.QtWidgets.QPushButton object at 0x000002880EA21CA0>, <PyQt5.QtWidgets.QTableWidget object at 0x000002880EA21820>, <PyQt5.QtWidgets.QLabel object at 0x0000028810BB8280>, <PyQt5.QtWidgets.QPushButton object at 0x0000028810BB8700>, <PyQt5.QtWidgets.QPushButton object at 0x0000028810BBBCA0>, <PyQt5.QtWidgets.QScrollBar object at 0x0000028812D8DAF0>, <PyQt5.QtWidgets.QPushButton object at 0x00000288130894C0>, <PyQt5.QtWidgets.QPushButton object at 0x0000028810BBB430>, <PyQt5.QtWidgets.QScrollBar object at 0x0000028810BB8C10>, <PyQt5.QtWidgets.QCheckBox object at 0x000002880EA21AF0>, <PyQt5.QtWidgets.QPushButton object at 0x0000028810BBBC10>, <PyQt5.QtWidgets.QListView object at 0x0000028812D8DB80>, <PyQt5.QtWidgets.QWidget object at 0x0000028812D8DC10>, <PyQt5.QtWidgets.QLabel object at 0x0000028810BB5EE0>, <PyQt5.QtWidgets.QPushButton object at 0x0000028810BB5700>, <PyQt5.QtWidgets.QWidget object at 0x0000028812D8DCA0>, <PyQt5.QtWidgets.QScrollBar object at 0x0000028812D8DD30>, <PyQt5.QtWidgets.QComboBox object at 0x000002880EA00D30>, <PyQt5.QtWidgets.QComboBox object at 0x0000028810BB8940>, <PyQt5.QtWidgets.QLineEdit object at 0x000002880EA21670>, <PyQt5.QtWidgets.QPushButton object at 0x0000028810BBB8B0>, <PyQt5.QtWidgets.QFrame object at 0x000002880EA211F0>, <PyQt5.QtWidgets.QListView object at 0x0000028812D8DDC0>, <PyQt5.QtWidgets.QLineEdit object at 0x0000028810BB8790>, <gui.widget_testplan.WidgetTestplan object at 0x0000028810BB84C0>, <PyQt5.QtWidgets.QLabel object at 0x0000028810BC10D0>, <PyQt5.QtWidgets.QWidget object at 0x0000028812D8DE50>, <PyQt5.QtWidgets.QLabel object at 0x0000028810BC11F0>, <PyQt5.QtWidgets.QPushButton object at 0x0000028810BC13A0>, <PyQt5.QtWidgets.QLineEdit object at 0x0000028812D8DEE0>, <PyQt5.QtWidgets.QAbstractButton object at 0x0000028812D8DF70>, <PyQt5.QtWidgets.QLineEdit object at 0x000002880EA21B80>, <PyQt5.QtWidgets.QLabel object at 0x0000028810BB5F70>, <PyQt5.QtWidgets.QLabel object at 0x0000028810BB58B0>, <PyQt5.QtWidgets.QLabel object at 0x0000028810BB81F0>, <PyQt5.QtWidgets.QPushButton object at 0x0000028810BC1310>, <PyQt5.QtWidgets.QLineEdit object at 0x00000288130893A0>, <PyQt5.QtWidgets.QWidget object at 0x0000028812D84040>, <PyQt5.QtWidgets.QFrame object at 0x000002880EA21160>, <PyQt5.QtWidgets.QWidget object at 0x0000028812D840D0>, <gui.main_widget.WgtStand object at 0x000002880EA00E50>, <PyQt5.QtWidgets.QWidget object at 0x0000028812D84160>, <PyQt5.QtWidgets.QListView object at 0x0000028812D841F0>, <PyQt5.QtWidgets.QPushButton object at 0x0000028810BC1430>, <PyQt5.QtWidgets.QHeaderView object at 0x0000028812D84280>]]
            meth=[.allWindows]                      value=[[<PyQt5.QtGui.QWindow object at 0x0000028811F158B0>, <PyQt5.QtGui.QWindow object at 0x0000028811F15820>, <PyQt5.QtGui.QWindow object at 0x0000028811F15670>, <PyQt5.QtGui.QWindow object at 0x0000028811F15700>, <PyQt5.QtGui.QWindow object at 0x0000028811F159D0>, <PyQt5.QtGui.QWindow object at 0x0000028811F15A60>]]
            meth=[.applicationDirPath]              value=[C:/Python3810]
            meth=[.applicationDisplayName]          value=[python]
            meth=[.applicationDisplayNameChanged]   value=[<unbound PYQT_SIGNAL applicationDisplayNameChanged()>]
            meth=[.applicationFilePath]             value=[C:/Python3810/python.exe]
            meth=[.applicationName]                 value=[python]
            meth=[.applicationPid]                  value=[17904]
            meth=[.applicationState]                value=[4]
            meth=[.applicationStateChanged]         value=[<unbound PYQT_SIGNAL applicationStateChanged(Qt::ApplicationState)>]
            meth=[.applicationVersion]              value=[3.8.10150.1013]
            meth=[.arguments]                       value=[['C:/!_STARICHENKO-T8/!!!_GD_additional/_PROJECTS/dwdm_test_system/testsystem.pyw']]
            meth=[.autoSipEnabled]                  value=[<built-in function autoSipEnabled>]
            attr=[.beep]                            value=[***MISSED SPECIAL***]
            meth=[.blockSignals]                    value=[<built-in function blockSignals>]
            meth=[.changeOverrideCursor]            value=[<built-in function changeOverrideCursor>]
            meth=[.childEvent]                      value=[<built-in function childEvent>]
            meth=[.children]                        value=[<built-in function children>]
            meth=[.clipboard]                       value=[<PyQt5.QtGui.QClipboard object at 0x0000028813089D30>]
            attr=[.closeAllWindows]                 value=[***MISSED DANGER***]
            meth=[.closingDown]                     value=[False]
            meth=[.colorSpec]                       value=[0]
            meth=[.commitDataRequest]               value=[<unbound PYQT_SIGNAL commitDataRequest(QSessionManager&)>]
            meth=[.connectNotify]                   value=[<built-in function connectNotify>]
            meth=[.cursorFlashTime]                 value=[1060]
            meth=[.customEvent]                     value=[<built-in function customEvent>]
            attr=[.deleteLater]                     value=[***MISSED DANGER***]
            meth=[.desktop]                         value=[<PyQt5.QtWidgets.QDesktopWidget object at 0x0000028813089D30>]
            meth=[.desktopFileName]                 value=[]
            meth=[.desktopSettingsAware]            value=[True]
            meth=[.destroyed]                       value=[<unbound PYQT_SIGNAL destroyed(QObject*)>]
            meth=[.devicePixelRatio]                value=[<built-in function devicePixelRatio>]
            meth=[.disconnect]                      value=[<built-in function disconnect>]
            meth=[.disconnectNotify]                value=[<built-in function disconnectNotify>]
            meth=[.doubleClickInterval]             value=[500]
            meth=[.dumpObjectInfo]                  value=[<built-in function dumpObjectInfo>]
            meth=[.dumpObjectTree]                  value=[<built-in function dumpObjectTree>]
            meth=[.dynamicPropertyNames]            value=[<built-in function dynamicPropertyNames>]
            meth=[.event]                           value=[<built-in function event>]
            meth=[.eventDispatcher]                 value=[<PyQt5.QtCore.QAbstractEventDispatcher object at 0x0000028813089DC0>]
            meth=[.eventFilter]                     value=[<built-in function eventFilter>]
            attr=[.exec]                            value=[***MISSED SPECIAL***]
            attr=[.exec_]                           value=[***MISSED SPECIAL***]
            attr=[.exit]                            value=[***MISSED DANGER***]
            meth=[.findChild]                       value=[<built-in function findChild>]
            meth=[.findChildren]                    value=[<built-in function findChildren>]
            meth=[.flush]                           value=[None]
            meth=[.focusChanged]                    value=[<unbound PYQT_SIGNAL focusChanged(QWidget*,QWidget*)>]
            meth=[.focusObject]                     value=[<PyQt5.QtWidgets.QPushButton object at 0x0000028810BBBCA0>]
            meth=[.focusObjectChanged]              value=[<unbound PYQT_SIGNAL focusObjectChanged(QObject*)>]
            meth=[.focusWidget]                     value=[<PyQt5.QtWidgets.QPushButton object at 0x0000028810BBBCA0>]
            meth=[.focusWindow]                     value=[<PyQt5.QtGui.QWindow object at 0x0000028813089DC0>]
            meth=[.focusWindowChanged]              value=[<unbound PYQT_SIGNAL focusWindowChanged(QWindow*)>]
            meth=[.font]                            value=[<PyQt5.QtGui.QFont object at 0x0000028811F389E0>]
            meth=[.fontChanged]                     value=[<unbound PYQT_SIGNAL fontChanged(QFont)>]
            meth=[.fontDatabaseChanged]             value=[<unbound PYQT_SIGNAL fontDatabaseChanged()>]
            attr=[.fontMetrics]                     value=[***MISSED SPECIAL***]
            meth=[.globalStrut]                     value=[PyQt5.QtCore.QSize()]
            meth=[.hasPendingEvents]                value=[False]
            meth=[.highDpiScaleFactorRoundingPolicy]value=[1]
            meth=[.inherits]                        value=[<built-in function inherits>]
            meth=[.inputMethod]                     value=[<PyQt5.QtGui.QInputMethod object at 0x0000028813089D30>]
            meth=[.installEventFilter]              value=[<built-in function installEventFilter>]
            meth=[.installNativeEventFilter]        value=[<built-in function installNativeEventFilter>]
            meth=[.installTranslator]               value=[<built-in function installTranslator>]
            meth=[.instance]                        value=[<PyQt5.QtWidgets.QApplication object at 0x000002887B44D430>]
            meth=[.isEffectEnabled]                 value=[<built-in function isEffectEnabled>]
            meth=[.isFallbackSessionManagementEnabled]value=[True]
            meth=[.isLeftToRight]                   value=[True]
            meth=[.isQuitLockEnabled]               value=[True]
            meth=[.isRightToLeft]                   value=[False]
            meth=[.isSavingSession]                 value=[<built-in function isSavingSession>]
            meth=[.isSessionRestored]               value=[<built-in function isSessionRestored>]
            meth=[.isSetuidAllowed]                 value=[False]
            meth=[.isSignalConnected]               value=[<built-in function isSignalConnected>]
            meth=[.isWidgetType]                    value=[<built-in function isWidgetType>]
            meth=[.isWindowType]                    value=[<built-in function isWindowType>]
            meth=[.keyboardInputInterval]           value=[400]
            meth=[.keyboardModifiers]               value=[<PyQt5.QtCore.Qt.KeyboardModifiers object at 0x0000028811F389E0>]
            meth=[.killTimer]                       value=[<built-in function killTimer>]
            meth=[.lastWindowClosed]                value=[<unbound PYQT_SIGNAL lastWindowClosed()>]
            meth=[.layoutDirection]                 value=[0]
            meth=[.layoutDirectionChanged]          value=[<unbound PYQT_SIGNAL layoutDirectionChanged(Qt::LayoutDirection)>]
            meth=[.libraryPaths]                    value=[['C:/Python3810/lib/site-packages/PyQt5/Qt5/plugins', 'C:/Python3810']]
            meth=[.metaObject]                      value=[<built-in function metaObject>]
            meth=[.modalWindow]                     value=[None]
            meth=[.mouseButtons]                    value=[<PyQt5.QtCore.Qt.MouseButtons object at 0x0000028811F38970>]
            attr=[.moveToThread]                    value=[***MISSED DANGER***]
            meth=[.notify]                          value=[<built-in function notify>]
            meth=[.objectName]                      value=[<built-in function objectName>]
            meth=[.objectNameChanged]               value=[<unbound PYQT_SIGNAL objectNameChanged(QString)>]
            meth=[.organizationDomain]              value=[]
            meth=[.organizationName]                value=[]
            meth=[.overrideCursor]                  value=[None]
            meth=[.palette]                         value=[<PyQt5.QtGui.QPalette object at 0x0000028811F38970>]
            meth=[.paletteChanged]                  value=[<unbound PYQT_SIGNAL paletteChanged(QPalette)>]
            meth=[.parent]                          value=[<built-in function parent>]
            meth=[.platformName]                    value=[windows]
            meth=[.postEvent]                       value=[<built-in function postEvent>]
            meth=[.primaryScreen]                   value=[<PyQt5.QtGui.QScreen object at 0x0000028813089DC0>]
            meth=[.primaryScreenChanged]            value=[<unbound PYQT_SIGNAL primaryScreenChanged(QScreen*)>]
            meth=[.processEvents]                   value=[None]
            meth=[.property]                        value=[<built-in function property>]
            attr=[.pyqtConfigure]                   value=[***MISSED SPECIAL***]
            meth=[.queryKeyboardModifiers]          value=[<PyQt5.QtCore.Qt.KeyboardModifiers object at 0x0000028811F38890>]
            meth=[.quit]                            value=[None]
            meth=[.quitOnLastWindowClosed]          value=[True]
            meth=[.receivers]                       value=[<built-in function receivers>]
            attr=[.removeEventFilter]               value=[***MISSED DANGER***]
            attr=[.removeLibraryPath]               value=[***MISSED DANGER***]
            attr=[.removeNativeEventFilter]         value=[***MISSED DANGER***]
            attr=[.removePostedEvents]              value=[***MISSED DANGER***]
            attr=[.removeTranslator]                value=[***MISSED DANGER***]
            meth=[.restoreOverrideCursor]           value=[None]
            meth=[.saveStateRequest]                value=[<unbound PYQT_SIGNAL saveStateRequest(QSessionManager&)>]
            meth=[.screenAdded]                     value=[<unbound PYQT_SIGNAL screenAdded(QScreen*)>]
            meth=[.screenAt]                        value=[<built-in function screenAt>]
            attr=[.screenRemoved]                   value=[***MISSED DANGER***]
            meth=[.screens]                         value=[[<PyQt5.QtGui.QScreen object at 0x000002880EA21430>, <PyQt5.QtGui.QScreen object at 0x000002880EA210D0>]]
            meth=[.sendEvent]                       value=[<built-in function sendEvent>]
            meth=[.sendPostedEvents]                value=[None]
            meth=[.sender]                          value=[<built-in function sender>]
            meth=[.senderSignalIndex]               value=[<built-in function senderSignalIndex>]
            meth=[.sessionId]                       value=[<built-in function sessionId>]
            meth=[.sessionKey]                      value=[<built-in function sessionKey>]
            attr=[.setActiveWindow]                 value=[***MISSED DANGER***]
            attr=[.setApplicationDisplayName]       value=[***MISSED DANGER***]
            attr=[.setApplicationName]              value=[***MISSED DANGER***]
            attr=[.setApplicationVersion]           value=[***MISSED DANGER***]
            attr=[.setAttribute]                    value=[***MISSED DANGER***]
            attr=[.setAutoSipEnabled]               value=[***MISSED DANGER***]
            attr=[.setColorSpec]                    value=[***MISSED DANGER***]
            attr=[.setCursorFlashTime]              value=[***MISSED DANGER***]
            attr=[.setDesktopFileName]              value=[***MISSED DANGER***]
            attr=[.setDesktopSettingsAware]         value=[***MISSED DANGER***]
            attr=[.setDoubleClickInterval]          value=[***MISSED DANGER***]
            attr=[.setEffectEnabled]                value=[***MISSED DANGER***]
            attr=[.setEventDispatcher]              value=[***MISSED DANGER***]
            attr=[.setFallbackSessionManagementEnabled]value=[***MISSED DANGER***]
            attr=[.setFont]                         value=[***MISSED DANGER***]
            attr=[.setGlobalStrut]                  value=[***MISSED DANGER***]
            attr=[.setHighDpiScaleFactorRoundingPolicy]value=[***MISSED DANGER***]
            attr=[.setKeyboardInputInterval]        value=[***MISSED DANGER***]
            attr=[.setLayoutDirection]              value=[***MISSED DANGER***]
            attr=[.setLibraryPaths]                 value=[***MISSED DANGER***]
            attr=[.setObjectName]                   value=[***MISSED DANGER***]
            attr=[.setOrganizationDomain]           value=[***MISSED DANGER***]
            attr=[.setOrganizationName]             value=[***MISSED DANGER***]
            attr=[.setOverrideCursor]               value=[***MISSED DANGER***]
            attr=[.setPalette]                      value=[***MISSED DANGER***]
            attr=[.setParent]                       value=[***MISSED DANGER***]
            attr=[.setProperty]                     value=[***MISSED DANGER***]
            attr=[.setQuitLockEnabled]              value=[***MISSED DANGER***]
            attr=[.setQuitOnLastWindowClosed]       value=[***MISSED DANGER***]
            attr=[.setSetuidAllowed]                value=[***MISSED DANGER***]
            attr=[.setStartDragDistance]            value=[***MISSED DANGER***]
            attr=[.setStartDragTime]                value=[***MISSED DANGER***]
            attr=[.setStyle]                        value=[***MISSED DANGER***]
            attr=[.setStyleSheet]                   value=[***MISSED DANGER***]
            attr=[.setWheelScrollLines]             value=[***MISSED DANGER***]
            attr=[.setWindowIcon]                   value=[***MISSED DANGER***]
            meth=[.signalsBlocked]                  value=[<built-in function signalsBlocked>]
            meth=[.startDragDistance]               value=[10]
            meth=[.startDragTime]                   value=[500]
            meth=[.startTimer]                      value=[<built-in function startTimer>]
            meth=[.startingUp]                      value=[False]
            obj=[.staticMetaObject]                 value=[<PyQt5.QtCore.QMetaObject object at 0x0000028811F389E0>]
            meth=[.style]                           value=[<PyQt5.QtWidgets.QCommonStyle object at 0x0000028810BB5EE0>]
            meth=[.styleHints]                      value=[<PyQt5.QtGui.QStyleHints object at 0x0000028810BB5EE0>]
            meth=[.styleSheet]                      value=[<built-in function styleSheet>]
            meth=[.sync]                            value=[None]
            meth=[.testAttribute]                   value=[<built-in function testAttribute>]
            meth=[.thread]                          value=[<built-in function thread>]
            meth=[.timerEvent]                      value=[<built-in function timerEvent>]
            meth=[.topLevelAt]                      value=[<built-in function topLevelAt>]
            meth=[.topLevelWidgets]                 value=[[<PyQt5.QtWidgets.QFrame object at 0x0000028810BBB940>, <PyQt5.QtWidgets.QWidget object at 0x0000028810BB5EE0>, <PyQt5.QtWidgets.QPushButton object at 0x0000028810BBB9D0>, <PyQt5.QtWidgets.QFrame object at 0x0000028810BBBB80>, <PyQt5.QtWidgets.QLabel object at 0x0000028810BB5DC0>, <PyQt5.QtWidgets.QPushButton object at 0x0000028810BBBA60>, <gui.authorization_widget.WidgetServerAuthorization object at 0x000002881306CF70>]]
            meth=[.topLevelWindows]                 value=[[<PyQt5.QtGui.QWindow object at 0x0000028810BB5EE0>, <PyQt5.QtGui.QWindow object at 0x0000028810BB5DC0>]]
            meth=[.tr]                              value=[<built-in function tr>]
            meth=[.translate]                       value=[<built-in function translate>]
            meth=[.wheelScrollLines]                value=[3]
            meth=[.widgetAt]                        value=[<built-in function widgetAt>]
            meth=[.windowIcon]                      value=[<PyQt5.QtGui.QIcon object at 0x0000028810BB5EE0>]
            ////////////////////////////////////////////////////////////////////////////////////////////////////

            Process finished with exit code -1073740940 (0xC0000374)
        """
        top_level_widgets = QApplication.topLevelWidgets()
        if not top_level_widgets:
            self.app = QApplication([])
            # QApplication.setQuitOnLastWindowClosed(False)     # dont help!

        # elif top_level_widgets and not QApplication.hasPendingEvents():
        #     msg = f"DETECTED START [{self.__class__.__name__=}] NOT IN MAIN THREAD!!!"
        #     UFU.logging_and_print_debug(msg)
        #     # raise Exception(msg)
        #     # QApplication.applicationName() - on child thread will show not blank!!!
        #     # QApplication.hasPendingEvents() - on child thread will show False!!!
        #     # for wgt_i in top_level_widgets:
        #     #     print(f"{wgt_i=}\t{type(wgt_i)=}\t{wgt_i.objectName()=}\t{wgt_i.accessibleName()=}")
        #
        #     # for wgt_i in top_level_widgets:
        #     #         __class__.WGT_ROOT = wgt_i.dialog_obj
        #
        #     if __class__.WGT_ROOT is not None:
        #         msg = f"EXISTS WGT_ROOT!!!"
        #         UFU.logging_and_print_warning(msg)
        #     else:
        #         msg = f"NOT FOUND WGT_ROOT!!!"
        #         UFU.logging_and_print_warning(msg)

    # =================================================================================================================
    @classmethod
    def last_answer_set(cls, result):
        cls.ANSWER_LAST = result
        cls.ANSWER_READY = True

    # SIMPLE ----------------------------------------------------------------------------------------------------------
    def _only_simple_0_nolock_noanswer(self, msg: Union[str, list[str]] = "msg", title: str = "title"):
        if isinstance(msg, (list, tuple)):   # msg
            msg = '\n'.join(msg)

        wgt = QFrame()
        wgt.setWindowFlags(Qt.ToolTip)

        wgt.setWindowTitle(title)
        wgt.setStyleSheet(f"background: rgb{COLOR_TUPLE_RGB.LIGHT_YELLOW}")
        wgt.setFrameStyle(3)
        wgt.setLineWidth(5)
        wgt.setMinimumWidth(300)
        wgt.setMinimumHeight(100)

        lbl = QLabel(msg)

        btn_1 = QPushButton("прочитано")
        btn_1.setStyleSheet("background: rgb(255,255,50)")
        btn_1.clicked.connect(wgt.close)

        layout = QVBoxLayout()
        layout.addWidget(lbl)
        layout.addWidget(btn_1)
        wgt.setLayout(layout)

        __class__.WGT_NOLOCK_NOANSWER = wgt
        __class__.WGT_NOLOCK_NOANSWER.show()

    def _only_simple_1_info(self, msg="msg", title="title"):
        answer = QMessageBox.information(self, title, msg, QMessageBox.Yes | QMessageBox.No)
        result = answer in [QMessageBox.Yes, 0]       # 0 если ACCEPTED! like wgt_i.accept()
        # UFU.logging_and_print_info(f"{answer=}")
        self.last_answer_set(result)
        return result

    def _only_simple_2_question(self, msg="msg", title="title"):
        answer = QMessageBox.question(self, title, msg, QMessageBox.Yes | QMessageBox.No)
        result = answer in [QMessageBox.Yes, 0]
        self.last_answer_set(result)
        return result

    def _only_simple_22_warning(self, msg="msg", title="title"):
        pass

    def _only_simple_3_critical(self, msg="msg", title="title"):
        answer = QMessageBox.critical(self, title, msg, QMessageBox.Yes | QMessageBox.No)
        result = answer in [QMessageBox.Yes, 0]
        self.last_answer_set(result)
        return result

    # INPUT -----------------------------------------------------------------------------------------------------------
    def _only_input_string(self, msg="Input string:", title="InputDialog"):
        result = None
        text, ok = QInputDialog.getText(self, title, msg)

        if ok and text:
            result = text

        self.last_answer_set(result)
        return result

    # SELECTORS -------------------------------------------------------------------------------------------------------
    def _only_select_file(self, title="SelectFile", path=''):
        """
        :param path:
            ""      - __file__
            "/"     - C:/
            ".."    - __file__/../
        """
        file_name = QFileDialog.getOpenFileName(self, title, path)[0]
        result = file_name or None

        self.last_answer_set(result)
        return result

    def _only_select_font(self, *args, **kwargs):
        # todo: finish!!!!
        result = None
        font_obj, ok = QFontDialog.getFont()
        if ok:
            result = str(font_obj)

        self.last_answer_set(result)
        return result

    def _only_select_color(self, *args, **kwargs):
        result = None
        color_obj = QColorDialog.getColor()
        if color_obj.isValid():
            result = color_obj.name()

        self.last_answer_set(result)
        return result

    # EXTENDED ========================================================================================================
    def _extended_thread_save_dialog(
            self,
            param1="msg",   # "msg"
            param2=None,      # title or path!
            _type=1,
            return_always_true=False,

            ask_only_if=True,
            func_link_do_while=None,

            autoaccept_tracking_value_link=None,
            autoaccept_link=None,
            timeout=0):
        """
        makes all dialogs thread safe!!!

        Получает ответ оператора - выдает True, если нажали OK иначе False.
        RECOMMENDED: use all parameters as KWARGS!!!

        :param ask_only_if: глобальный вопрос - показывать или нет окно вопроса - если нет то результатом функции будет True чтобы не мешать дальнейшим процессам!!!
        :param func_link_do_while: если указана - перед отображением окна она проверяется и окно протолжает отображаться то тех пор пока не будет изменено значение на False!
            или нажата кнопка NO - РАБОТАЕТ!!!
        :param return_always_true: создана для случая когда нам не важен ответ, но и нежелательно выдавать False! всегда будет выдавать True

        :param autoaccept_tracking_value_link: функция, если указана, если указано, то в окно вместе с сообщением выводится полученое значение (и обновляется с прохождением autoaccept_link)
        :param autoaccept_link: функция, если указана, то проверяется c периодичностью, если получила результат True - то считается что нажата кнопка OK, окно закрывается!
        :param timeout: время в секундах - если задана функция autoaccept_link, то она проверяется именно с этим промежутком времени

        How to use text formatting in gui ask:
            <br><br> - brake line
            <b>ПРЕДУПРЕЖДЕНИЕ</b> - Bold
            <font size=20 color=red></font> - font size/ color not worked!
        """
        # INIT RESULTS ------------------------------------------------------------------------------------------------
        __class__.ANSWER_READY = None
        __class__.ANSWER_LAST = None
        __class__.ANSWER_LAST_WAS_AUTOACCEPTED = None

        # -------------------------------------------------------------------------------------------------------------
        # INPUT
        if isinstance(param1, (list, tuple)):   # msg
            param1 = '\n'.join(param1)

        if func_link_do_while is None:
            func_link_do_while = LAMBDA_TRUE
        elif not callable(func_link_do_while):
            msg = f"получено неверное значение для [{func_link_do_while=}]"
            print(msg)
            raise Exception(msg)

        # -------------------------------------------------------------------------------------------------------------
        if __class__.WGT_ROOT is None:
            if _type == 0:
                emit_link = self.s_type0.emit
            elif _type == 1:
                emit_link = self.s_type1.emit
            elif _type == 2:
                emit_link = self.s_type2.emit
            elif _type == 3:
                emit_link = self.s_type3.emit
            elif _type == 4:
                emit_link = self.s_type4.emit
            elif _type == 5:
                emit_link = self.s_type5.emit
            elif _type == 6:
                emit_link = self.s_type6.emit
            elif _type == 7:
                emit_link = self.s_type7.emit

        else:
            if _type == 0:
                emit_link = __class__.WGT_ROOT.s_type0.emit
            elif _type == 1:
                emit_link = __class__.WGT_ROOT.s_type1.emit
            elif _type == 2:
                emit_link = __class__.WGT_ROOT.s_type2.emit
            elif _type == 3:
                emit_link = __class__.WGT_ROOT.s_type3.emit
            elif _type == 4:
                emit_link = __class__.WGT_ROOT.s_type4.emit
            elif _type == 5:
                emit_link = __class__.WGT_ROOT.s_type5.emit
            elif _type == 6:
                emit_link = __class__.WGT_ROOT.s_type6.emit
            elif _type == 7:
                emit_link = __class__.WGT_ROOT.s_type7.emit

        # -------------------------------------------------------------------------------------------------------------
        result = False
        if ask_only_if:
            time_start = time.time()
            while func_link_do_while():
                # timeout ------------------------------------------------------------------------
                if timeout and time.time() - time_start > timeout:
                    break

                # start thread to close Dialog!!! ------------------------------------------------
                if autoaccept_link is not None:
                    self.autoaccept_start_thread(autoaccept_link)

                # start Dialog!!! ------------------------------------------------
                emit_link(param1, param2)

                if _type == 0:
                    break

                while True:
                    print(f"ожидание ответа пользователя [{autoaccept_link=}/{__class__.ANSWER_LAST_WAS_AUTOACCEPTED=}]")
                    if __class__.ANSWER_READY:
                        result = __class__.ANSWER_LAST
                        break
                    elif __class__.ANSWER_LAST_WAS_AUTOACCEPTED:
                        result = True
                        break
                    time.sleep(1)

                _msg = f"Интерактивный ответ пользователя [{result=}/{param1=}/{param2=}]"
                print(_msg, result)

                if not result or func_link_do_while == LAMBDA_TRUE:
                    break

                time.sleep(0.5)

        if return_always_true:
            result = True
        return result

    # SIMPLE ----------------------------------------------------------------------------------------------------------
    def simple_0_nolock_noanswer(self, *args, **kwargs):
        return self._extended_thread_save_dialog(*args, _type=0, **kwargs)

    def simple_1_info(self, *args, **kwargs):
        return self._extended_thread_save_dialog(*args, _type=1, **kwargs)

    def simple_2_question(self, *args, **kwargs):
        return self._extended_thread_save_dialog(*args, _type=2, **kwargs)

    def simple_3_critical(self, *args, **kwargs):
        return self._extended_thread_save_dialog(*args, _type=3, **kwargs)

    # INPUT -----------------------------------------------------------------------------------------------------------
    def input_string(self, *args, **kwargs):
        return self._extended_thread_save_dialog(*args, _type=4, **kwargs)

    # SELECTORS -------------------------------------------------------------------------------------------------------
    def select_file(self, *args, **kwargs):
        return self._extended_thread_save_dialog(*args, _type=5, **kwargs)

    def select_font(self, *args, **kwargs):
        return self._extended_thread_save_dialog(*args, _type=6, **kwargs)

    def select_color(self, *args, **kwargs):
        return self._extended_thread_save_dialog(*args, _type=7, **kwargs)

    # COMPOSITION =====================================================================================================
    def _composed_box(self, msg="button_box?", title="button_box"):
        """
        composed msg box by wgts!
        if you need simple ask YesNo - use other simple from the beginning!
        """
        qdialog = QDialog()

        layout = QVBoxLayout(qdialog)
        layout.addWidget(QLabel(msg))

        buttons = QDialogButtonBox(qdialog)
        buttons.setStandardButtons(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        # buttons.setOrientation(Qt.Horizontal)
        layout.addWidget(buttons)

        qdialog.setWindowTitle(title)

        buttons.accepted.connect(qdialog.accept)
        buttons.rejected.connect(qdialog.reject)
        qdialog.exec_()
        result = qdialog.result() == 1    # 0=Cancel, 1=OK
        self.last_answer_set(result)
        return result

    # =================================================================================================================
    def _QDialog(self):
        qdialog = QDialog()
        qdialog.exec_()

    def _test_some(self):
        self.simple_1_info("hello1")
        self.simple_2_question("hello2")
        self.simple_3_critical("hello3")
        self.input_string("hello4")
        self.select_file("hello5")
        self.select_font("hello6")
        self.select_color("hello7")


# =====================================================================================================================
INSTRUCTION_HTML = """
    <html>
    #########################################################################
    <h1>ИНСТРУКЦИЯ ПО ТЕСТИРОВАНИЮ BACKPLANE (ШАССИ КРЕЙТА)</h1>
    
    <h2>1. В ШАССИ УСТАНОВИТЕ ВСЕ УСТРОЙСТВА!</h2>
    <p>
    - Не должно быть ни одного свободного места (все PS, все CU и FU должны быть заняты)<br>
    </p>
    
    <h2>2. CU</h2>
    <p>
    - могут быть разными, но должна быть одинаковая версия ATLAS!<br>
    - Версия Atlas должна быть не ниже 2.3.5 (иначе будут ошибки проверки ETHERNET)<br>
    - Запуск необходимо производить с CU0, а не с CU1<br>
    (иначе будут ошибки в определении подключенных устройств)<br>
    </p>
    
    <h2>3. устройства В8</h2>
    <p>
    - во все слоты B8 установите КАМА не ниже v1.1<br>
    - если будут отказы на КАМА, то попробуйте ревизию плат v1.3<br>
    - если использовать НЕ КАМА, то могут не пройти тесты ETHERNET!<br>
    - если использовать ПАССИВНЫЕ УСТРОЙСТВА, то не пройдут тесты UART и ETHERNET!<br>
    </p>
    
    <h2>4. БЛОКИ ПИТАНИЯ</h2>
    <p>
    - могут быть любые разные (даже AC/DC)<br>
    - Подайте питание на оба блока питания<br>
    </p>
    
    <h2>5. Запустите АвтоТестирование и следуйте указаниям</h2>
    
    <h3>ПРИМЕЧАНИЕ:</h3>
    <p>
    - если будут ошибки когда их очевидно не должно быть - перезапустите VOLGA!
    </p>
    
    <p>
    <b>YES/NO</b>=Прочинано - продолжить работу
    </p>
    
    </html>
"""


def try__autoaccept():
    GuiDialog().simple_1_info(INSTRUCTION_HTML, autoaccept_link=Sleep.TRUE)
    GuiDialog().simple_1_info(INSTRUCTION_HTML, autoaccept_link=Sleep.TRUE)


def try__GuiDialog():
    GuiDialog().simple_1_info(INSTRUCTION_HTML)
    GuiDialog().simple_2_question('<h1>hel\nlo</h1><b>12345<br>1234</b> 21211212 <br>')
    GuiDialog().simple_3_critical()
    GuiDialog().input_string()
    GuiDialog().select_file()
    GuiDialog().select_font()
    GuiDialog().select_color()
    GuiDialog()._composed_box()
    GuiDialog()._QDialog()
    GuiDialog().simple_1_info(INSTRUCTION_HTML)


def try__static():
    print(GuiDialog().simple_1_info(INSTRUCTION_HTML))
    print(GuiDialog().simple_1_info(INSTRUCTION_HTML))


# =====================================================================================================================
if __name__ == '__main__':
    try__info()
    # try__autoaccept()
    # print(GuiDialog().simple_1_info('<h1>hel\nlo</h1><b>12345<br>1234</b> 21211212 <br>'))
    # print(GuiDialog().simple_1_info('<h1>hel\nlo</h1><b>12345<br>1234</b> 21211212 <br>'))


# =====================================================================================================================



from base_aux.base_lambdas.m1_lambda import *

from base_aux.base_enums.m2_enum1_adj import *
from base_aux.base_values.m2_value_special import NoValue


# =====================================================================================================================
class VictimStd(Enum):
    NONE = None
    A1 = 1
    TUPLE = (1, 2)
    STR_LOWER = "str_lower"
    STR_UPPER = "STR_UPPER"


class VictimEq(NestEq_EnumAdj):
    NONE = None
    A1 = 1
    TUPLE = (1, 2)
    STR_LOWER = "str_lower"
    STR_UPPER = "STR_UPPER"


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="source, _EXPECTED",
    argvalues=[
        (None, None),
        (1, 1),
        ((1, 2), (1, 2)),
        ((1, 22), NoValue),
        ("str_lower", "str_lower"),
        ("STR_LOWER", "str_lower"),
    ]
)
def test___value__get_original(source, _EXPECTED):
    Lambda(VictimEq._value__get_original, source).check_expected__assert(_EXPECTED)

# ---------------------------------------------------------------------------------------------------------------------
@pytest.mark.parametrize(
    argnames="source, _EXPECTED",
    argvalues=[
        (None, VictimEq.NONE),
        (1, VictimEq.A1),
        ((1, 2), VictimEq.TUPLE),
        ((1, 22), Exception),
        ("str_lower", VictimEq.STR_LOWER),
        # ("STR_LOWER", VictimEq.STR_LOWER),    # FIXME: CANT CREATE!
    ]
)
def test__init(source, _EXPECTED):
    Lambda(VictimEq, source).check_expected__assert(_EXPECTED)


# =====================================================================================================================
class Test_EnumStd:
    @pytest.mark.parametrize(
        argnames="source, other, _EXPECTED",
        argvalues=[
            # NONE --------
            (VictimStd, None, (False, True)),
            (VictimStd, VictimStd(None), (False, True)),
            (VictimStd, VictimStd.NONE, (False, True)),

            (VictimStd(None), None, (False, Exception)),
            (VictimStd(None), VictimStd.NONE, (True, Exception)),

            (VictimStd.NONE, None, (False, Exception)),
            (VictimStd.NONE, VictimStd(None), (True, Exception)),

            # 1 --------
            (VictimStd, 1, (False, True)),
            (VictimStd, VictimStd(1), (False, True)),
            (VictimStd, VictimStd.A1, (False, True)),

            (VictimStd(1), 1, (False, Exception)),
            (VictimStd(1), VictimStd.A1, (True, Exception)),

            (VictimStd.A1, 1, (False, Exception)),
            (VictimStd.A1, VictimStd(1), (True, Exception)),

            # DIFF --------
            (VictimStd.A1, VictimStd.NONE, (False, Exception)),
        ]
    )
    def test__eq_in(self, source, other, _EXPECTED):
        Lambda(source == other).check_expected__assert(_EXPECTED[0])

        func_link = lambda x: x in source
        Lambda(func_link, other).check_expected__assert(_EXPECTED[1])


# =====================================================================================================================
class Test_EnumEq:
    @pytest.mark.parametrize(
        argnames="source, other, _EXPECTED",
        argvalues=[
            # NONE --------
            (VictimEq, None, (False, True)),
            (VictimEq, VictimEq(None), (False, True)),
            (VictimEq, VictimEq.NONE, (False, True)),

            (VictimEq(None), None, (True, Exception)),
            (VictimEq(None), VictimEq.NONE, (True, Exception)),

            (VictimEq.NONE, None, (True, Exception)),
            (VictimEq.NONE, VictimEq(None), (True, Exception)),

            # 1 --------
            (VictimEq, 1, (False, True)),
            (VictimEq, VictimEq(1), (False, True)),
            (VictimEq, VictimEq.A1, (False, True)),

            (VictimEq(1), 1, (True, Exception)),
            (VictimEq(1), VictimEq.A1, (True, Exception)),

            (VictimEq.A1, 1, (True, Exception)),
            (VictimEq.A1, VictimEq(1), (True, Exception)),

            # TUPLE --------
            (VictimEq, (1,2), (False, True)),
            (VictimEq, VictimEq((1,2)), (False, True)),
            (VictimEq, VictimEq.TUPLE, (False, True)),

            (VictimEq((1,2)), (1,2), (True, Exception)),
            (VictimEq((1,2)), VictimEq.TUPLE, (True, Exception)),

            (VictimEq.TUPLE, (1,2), (True, Exception)),
            (VictimEq.TUPLE, VictimEq((1,2)), (True, Exception)),

            # STR --------
            (VictimEq, "str_lower", (False, True)),
            (VictimEq, "STR_LOWER", (False, False)),

            (VictimEq("str_lower"), "str_lower", (True, Exception)),
            (VictimEq("str_lower"), "STR_LOWER", (False, Exception)),

            # DIFF values = same cls--------
            (VictimEq.A1, VictimEq.NONE, (False, Exception)),

            # DIFF clss --------
            (VictimEq.NONE, VictimStd.NONE, (False, Exception)),
            (VictimEq, VictimStd, (False, False)),
            (VictimStd, VictimEq, (False, False)),
        ]
    )
    def test__eq_in(self, source, other, _EXPECTED):
        funk_link = lambda: source == other
        Lambda(funk_link).check_expected__assert(_EXPECTED[0])

        func_link = lambda x: x in source
        Lambda(func_link, other).check_expected__assert(_EXPECTED[1])


# =====================================================================================================================
def _examples() -> None:
    WHEN = EnumAdj_When2.BEFORE
    if WHEN is EnumAdj_When2.BEFORE:
        pass

    print(EnumAdj_NumFPoint.COMMA)  # EnumAdj_NumFPoint.COMMA
    print(EnumAdj_NumFPoint("."))  # EnumAdj_NumFPoint.DOT

    print("." in EnumAdj_NumFPoint)  # True
    print(EnumAdj_NumFPoint.DOT in EnumAdj_NumFPoint)  # True

    print(EnumAdj_NumFPoint(".") == ".")  # False
    print(EnumAdj_NumFPoint(EnumAdj_NumFPoint.DOT))  # EnumAdj_NumFPoint.DOT     # BEST WAY to init value!

    # ObjectInfo(VictimEq).print()

    # ITERATE
    print()
    for i in VictimEq:
        print(i, i.name, i.value)
        print()


# =====================================================================================================================
if __name__ == "__main__":
    _examples()


# =====================================================================================================================



from PyQt5.QtCore import QThread

from base_aux.base_lambdas.m1_lambda import *
from base_aux.base_lambdas.m2_lambda3_thread import LambdaThread
from base_aux.base_types.m0_static_typing import *


# =====================================================================================================================
TYPING__LAMBDA_LIST__DRAFT = list[Lambda | Callable | type | Any]
TYPING__LAMBDA_LIST__FINAL = list[Lambda]


# =====================================================================================================================
class LambdaListThread(QThread):
    """
    GOAL
    ----
    call all base_lambdas in list one by one + return there Lambda-objects in list
    results will kept in objects
    """
    LAMBDAS: TYPING__LAMBDA_LIST__FINAL
    PROCESS_ACTIVE: EnumAdj_ProcessStateActive = EnumAdj_ProcessStateActive.NONE

    def __init__(self, lambdas: TYPING__LAMBDA_LIST__DRAFT, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        result = []
        for item in lambdas:
            if not isinstance(item, Lambda):
                item = Lambda(item)
            result.append(item)
        self.LAMBDAS = result

    def run(self) -> None:
        # ONLY ONE EXECUTION on instance!!! dont use locks! -------------
        if self.PROCESS_ACTIVE == EnumAdj_ProcessStateActive.STARTED:
            return
        self.PROCESS_ACTIVE = EnumAdj_ProcessStateActive.STARTED

        # FIN ----------------------------------------------------------
        for item in self.LAMBDAS:
            if isinstance(item, LambdaThread):
                item.start()
            else:
                item.run()

        self.wait_finished__all()
        self.PROCESS_ACTIVE = EnumAdj_ProcessStateActive.FINISHED

    # OVERWRITE! ======================================================================================================
    def __call__(self, *args, **kwargs) -> TYPING__LAMBDA_LIST__FINAL:
        self.run()
        return self.LAMBDAS

    # =================================================================================================================
    def check_raise__any(self) -> bool:
        self.run()
        self.wait_finished()

        for item in self.LAMBDAS:
            if item.EXC is not None:
                return True
            else:
                return False

    def check_no_raise__any(self) -> bool:
        return not self.check_raise__any()

    def wait_finished__all(self, sleep: float = 1) -> None:
        if self.PROCESS_ACTIVE == EnumAdj_ProcessStateActive.NONE:
            self.run()

        for item in self.LAMBDAS:
            item.wait_finished(sleep)

    def wait_finished(self, sleep: float = 1) -> None:
        if self.PROCESS_ACTIVE == EnumAdj_ProcessStateActive.NONE:
            self.run()

        count = 1
        while self.PROCESS_ACTIVE != EnumAdj_ProcessStateActive.FINISHED:
            print(f"wait_finished {count=}")
            count += 1
            time.sleep(sleep)


# =====================================================================================================================
def _explore():
    pass


# =====================================================================================================================
if __name__ == "__main__":
    _explore()
    pass


# =====================================================================================================================



from base_aux.testplans.tc import *
from base_aux.valid.m3_valid_chains import *
from .tc0_groups import *


# =====================================================================================================================
class TestCase(Base_TestCase):
    ASYNC = True
    DESCRIPTION = "ptb"

    @classmethod
    def startup__cls__wrapped(cls) -> TYPING__RESULT_W_NORETURN:
        # return True
        result_chain = ValidChains(
            [
                Valid(value_link=hasattr(cls, "STAND"), name="hasattr STAND"),
                Valid(value_link=hasattr(cls.STAND, "DEV_LINES"), name="hasattr DEV_LINES"),
                Valid(value_link=hasattr(cls.STAND.DEV_LINES, "ATC"), name="hasattr ATC"),
                Valid(value_link=cls.STAND.DEV_LINES.ATC[0].connect, name="ATC.connect()"),
            ],
        )
        return result_chain

    def startup__wrapped(self) -> TYPING__RESULT_W_NORETURN:
        result = ValidChains(
            [
                Valid(value_link=self.DEV_COLUMN.DUT.connect__only_if_address_resolved, name="DUT.connect__only_if_address_resolved"),
            ],
        )
        return result

    def run__wrapped(self) -> TYPING__RESULT_W_NORETURN:
        # time.sleep(0.1)
        result = ValidChains(
            [
                Valid(value_link=self.DEV_COLUMN.DUT.connect__only_if_address_resolved, name="DUT.connect__only_if_address_resolved"),
            ],
        )
        return result


# =====================================================================================================================



from base_aux.aux_attr.m2_annot3_gacls_keys_as_values import NestGaCls_AnnotNamesAsValuesIc


# =====================================================================================================================
class FilesStdExtensions(NestGaCls_AnnotNamesAsValuesIc):
    """
    GOAL
    ----
    use as collection for standard extension

    SPECIALLY CREATED FOR
    ---------------------
    Resolve_DirPath for detecting files withing names which are not exist
    """

    # TEXT -------------
    TXT: str
    JSON: str
    CSV: str
    INI: str
    INF: str
    IML: str
    XML: str
    YML: str
    FLAKE8: str
    MD: str
    GITIGNORE: str
    LOG: str
    HTML: str
    REG: str

    # EXECUTIVE ---------
    PY: str
    PYC: str
    C: str
    CPP: str
    H: str
    BAT: str
    CMD: str
    SH: str
    EXE: str
    MSI: str

    # MEDIA -------------
    ACCDB: str

    # graf
    BMP: str
    JPG: str
    JPEG: str
    PNG: str
    GIF: str
    TIFF: str
    ICO: str
    RAW: str

    # sound
    MP3: str
    FLAC: str
    APE: str
    OGG: str
    WAW: str
    AC3: str
    WMA: str
    M4A: str
    AAC: str

    # video
    AVI: str
    WMW: str
    MKV: str
    # 3GP: str
    FLV: str
    MPEG: str
    MP4: str
    MOV: str
    VOB: str

    # EXTRA -------------
    LNK: str

    # arch
    BK: str
    ISO: str
    RAR: str
    ZIP: str
    # 7Z: str

    # app
    PDF: str
    DOC: str
    DOCX: str
    XLS: str
    XLSX: str
    PPT: str
    PPTX: str


# =====================================================================================================================



from typing import *


# =====================================================================================================================
class NestCall_MethodName:
    """
    GOAL
    ----
    separate specially as base with not exact method name

    SPECIALLY CREATED FOR
    ---------------------
    Base_Alert
    """
    _CALL__METHOD_NAME: str

    def __call__(self, *args, **kwargs) -> Any | NoReturn:
        method = getattr(self, self._CALL__METHOD_NAME)
        return method(*args, **kwargs)


# =====================================================================================================================
class NestCall_Resolve:
    """
    GOAL
    ----
    just show that nested class used basically for main purpose which will returned by .resolve() method
    its like an aux-function but with better handling

    NOTE
    ----
    dont use it as type!
    dont keep in attributes!
    resolve ti inline!

    SPECIALLY CREATED FOR
    ---------------------
    files.filepath
    """
    def __call__(self, *args, **kwargs) -> Any | NoReturn:
        return self.resolve(*args, **kwargs)

    def resolve(self, *args, **kwargs) -> Any | NoReturn:
        return NotImplemented


# =====================================================================================================================
class NestCall_Other:
    """
    GOAL
    ----
    apply other object as data

    SPECIALLY CREATED FOR
    ---------------------
    text formatted
    """
    def __call__(self, other: Any, *args, **kwargs) -> Any | NoReturn:
        return self.other(other, *args, **kwargs)

    def other(self, other: Any, *args, **kwargs) -> Any | NoReturn:
        return NotImplemented


# =====================================================================================================================



from base_aux.base_lambdas.m1_lambda import *
from base_aux.base_enums.m2_enum1_adj import EnumAdj_When2


# =====================================================================================================================
class Lambda_Bool(Lambda):
    """
    GOAL
    ----
    same as Lambda, in case of get result in bool variant
    +add reverse

    SPECIALLY CREATED FOR
    ---------------------
    classes.Valid.skip_link with Reverse variant

    why Reversing is so important?
    --------------------------------
    because you cant keep callable link and reversing it by simply NOT
        skip_link__direct = bool        # correct
        skip_link__direct = Lambda_Bool(bool)  # correct
        skip_link__reversal = not bool  # incorrect
        skip_link__reversal = Lambda_Bool(bool, attr).get_reverse  # correct

    but here we can use lambda
        skip_link__reversal = lambda attr: not bool(attr)  # correct but not so convenient ???

    PARAMS
    ------
    :ivar BOOL_REVERSE: just for Lambda_BoolReversed, no need to init
    """
    def resolve(self, *args, **kwargs) -> bool | NoReturn:
        self.run(*args, **kwargs)

        if self.EXC is not None:
            raise self.EXC
        else:
            return bool(self.RESULT)


# ---------------------------------------------------------------------------------------------------------------------
class Lambda_BoolReversed(Lambda_Bool):
    """
    just a reversed Lambda_Bool
    """
    def resolve(self, *args, **kwargs) -> bool | NoReturn:
        self.run(*args, **kwargs)

        if self.EXC is not None:
            raise self.EXC
        else:
            return not bool(self.RESULT)


# =====================================================================================================================
class Lambda_TrySuccess(Lambda_Bool):
    """
    just an ability to check if object is not raised on call

    BEST PRACTICE
    -------------
    1. direct/quick/shortest checks without big trySentence
        if Lambda_TrySuccess(func):
            return func()

    2. pytestSkipIf
        @pytest.mark.skipif(Lambda_TryFail(func), ...)

    3. pytest assertions

        class Victim(DictIcKeys_Ga_AnnotRequired):
            lowercase: str

        assert Lambda_TryFail(Victim)
        assert not Lambda_TrySuccess(Victim)
        assert Lambda_TrySuccess(Victim, lowercase="lowercase")

    EXAMPLES
    --------
        if callables and Lambda_TrySuccess(getattr, source, name) and callable(getattr(source, name)):
            continue

        so here raise is acceptable in getattr(source, name) in case of PROPERTY_RAISE
    """
    def resolve(self, *args, **kwargs) -> bool:
        self.run(*args, **kwargs)

        if self.EXC is not None:
            return False
        else:
            return True


# ---------------------------------------------------------------------------------------------------------------------
class Lambda_TryFail(Lambda_TrySuccess):
    def resolve(self, *args, **kwargs) -> bool:
        self.run(*args, **kwargs)

        if self.EXC is not None:
            return True
        else:
            return False


# =====================================================================================================================
class Lambda_Sleep(Lambda):
    """
    GOAL
    ----
    it is not just a sleep func!
    it sleep added for other func! - before or after its execution!
    """
    WHEN: EnumAdj_When2 = EnumAdj_When2.BEFORE
    SEC: float = 1

    def __init__(self, *args, sec: float = None, **kwargs) -> None:
        if sec is not None:
            self.SEC = sec
        super().__init__(*args, **kwargs)

    def resolve(self, sec: float = None, *args, **kwargs) -> Any | NoReturn:
        if sec is None:
            sec = self.SEC

        if self.WHEN is EnumAdj_When2.BEFORE:
            time.sleep(sec)

        self.run(*args, **kwargs)

        if self.WHEN is EnumAdj_When2.AFTER:
            time.sleep(sec)

        if self.EXC is not None:
            raise self.EXC
        else:
            return self.RESULT


# ---------------------------------------------------------------------------------------------------------------------
class Lambda_SleepAfter(Lambda_Sleep):
    """
    CREATED SPECIALLY FOR
    ---------------------
    UART/ATC tests for RST command
    """
    WHEN: EnumAdj_When2 = EnumAdj_When2.AFTER


# =====================================================================================================================



from base_aux.base_lambdas.m1_lambda import *

from base_aux.base_values.m4_primitives import *


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="source, args, _EXPECTED",
    argvalues=[
        (Exception, (), (False, True,
                         Exception, Exception, Exception, False, VALUE_SPECIAL.SKIPPED, Exception)),  # be careful here! exc in source return exc NO RAISE!!!!
        (Exception(), (), (False, True,
                Exception, Exception, Exception, False, Exception, Exception)),
        (LAMBDA_EXC, (), (False, True,
                          Exception, Exception, Exception, False, VALUE_SPECIAL.SKIPPED, Exception)),
        (LAMBDA_RAISE, (), (True, False,
                            Exception, None, Exception, False, VALUE_SPECIAL.SKIPPED, VALUE_SPECIAL.SKIPPED)),

        (LAMBDA_TRUE, (), (False, True,
                           True, True, True, True, VALUE_SPECIAL.SKIPPED, True)),
        (LAMBDA_FALSE, (), (False, True,
                            False, False, False, False, VALUE_SPECIAL.SKIPPED, False)),
        (LAMBDA_NONE, (), (False, True,
                           None, None, None, False, VALUE_SPECIAL.SKIPPED, None)),

        (True, (), (False, True,
                True, True, True, True, True, True)),
        (False, (), (False, True,
                False, False, False, False, False, False)),
        (None, (), (False, True,
                None, None, None, False, None, None)),

        (INST_CALL_TRUE, (), (False, True,
                              True, True, True, True, VALUE_SPECIAL.SKIPPED, True)),
        (INST_CALL_FALSE, (), (False, True,
                               False, False, False, False, VALUE_SPECIAL.SKIPPED, False)),
        (INST_CALL_RAISE, (), (True, False,
                               Exception, None, Exception, False, VALUE_SPECIAL.SKIPPED, VALUE_SPECIAL.SKIPPED)),

        (INST_BOOL_TRUE, (),  (False, True,
                INST_BOOL_TRUE, INST_BOOL_TRUE, INST_BOOL_TRUE, True, INST_BOOL_TRUE, INST_BOOL_TRUE)),
        (INST_BOOL_FALSE, (), (False, True,
                INST_BOOL_FALSE, INST_BOOL_FALSE, INST_BOOL_FALSE, False, INST_BOOL_FALSE, INST_BOOL_FALSE)),
        (INST_BOOL_RAISE, (), (False, True,
                INST_BOOL_RAISE, INST_BOOL_RAISE, INST_BOOL_RAISE, False, INST_BOOL_RAISE, INST_BOOL_RAISE)),

        # collections ----------
        ((), (), (False, True,
                (), (), (), False, (), ())),
        ([], (), (False, True,
                [], [], [], False, [], [])),
        (LAMBDA_LIST_KEYS, (), (False, True,
                                [], [], [], False, VALUE_SPECIAL.SKIPPED, [])),

        ([None, ], (), (False, True,
                [None, ], [None, ], [None, ], True, [None, ], [None, ])),
        ([1, ], (), (False, True,
                [1, ], [1, ], [1, ], True, [1, ], [1, ])),
    ]
)
def test__get_result(source, args, _EXPECTED):
    Lambda(Lambda(source, *args).check_raised__bool).check_expected__assert(_EXPECTED[0])
    Lambda(Lambda(source, *args).check_no_raised__bool).check_expected__assert(_EXPECTED[1])
    Lambda(Lambda(source, *args).resolve__raise).check_expected__assert(_EXPECTED[2])
    Lambda(Lambda(source, *args).resolve__raise_as_none).check_expected__assert(_EXPECTED[3])
    Lambda(Lambda(source, *args).resolve__exc).check_expected__assert(_EXPECTED[4])
    Lambda(Lambda(source, *args).resolve__bool).check_expected__assert(_EXPECTED[5])
    Lambda(Lambda(source, *args).resolve__skip_callables).check_expected__assert(_EXPECTED[6])
    Lambda(Lambda(source, *args).resolve__skip_raised).check_expected__assert(_EXPECTED[7])


# =====================================================================================================================



from base_aux.testplans.devices_base import *
from base_aux.testplans.tc import *


# =====================================================================================================================
class Victim_DevicesLines(DeviceKit):
    ATC: TableLine = TableLine(Base_Device)


# =====================================================================================================================
@pytest.mark.skip    # FIXME: ref or not!
class Test__TC:
    @classmethod
    def setup_class(cls):
        pass
        cls.Victim = Base_TestCase

    # @classmethod
    # def teardown_class(cls):
    #     pass
    #
    # def setup_method(self, method):
    #     passtest__tc.py
    #
    # def teardown_method(self, method):
    #     pass

    # -----------------------------------------------------------------------------------------------------------------
    def test__cls(self):
        # EXISTS IN CLS --------------
        assert len(self.Victim.TCSi_LINE) == 0
        assert self.Victim.STAND.DEV_LINES is None

        assert self.Victim.result__startup_cls is None
        assert self.Victim.result__teardown_cls is None

        # EXISTS IN INSTANCE --------------
        assert not hasattr(self.Victim, "INDEX")
        # assert not hasattr(self.Victim, "SETTINGS")
        assert hasattr(self.Victim, "DEV_COLUMN")

        assert not hasattr(self.Victim, "details")
        assert not hasattr(self.Victim, "exc")

    def test__cls__devices_apply__NONE(self):
        assert len(self.Victim.TCSi_LINE) == 0

        # EXISTS IN CLS --------------
        assert self.Victim.STAND.DEV_LINES is None

        assert self.Victim.result__startup_cls is None
        assert self.Victim.result__teardown_cls is None

        # EXISTS IN INSTANCE --------------
        assert not hasattr(self.Victim, "INDEX")
        # assert not hasattr(self.Victim, "SETTINGS")
        assert hasattr(self.Victim, "DEV_COLUMN")

        assert not hasattr(self.Victim, "details")
        assert not hasattr(self.Victim, "exc")

    def test__cls__devices_apply__example(self):
        assert len(self.Victim.TCSi_LINE) == 0
        self.Victim.STAND.DEV_LINES = Victim_DevicesLines()

        # EXISTS IN CLS --------------
        assert self.Victim.STAND.DEV_LINES is not None

        assert self.Victim.result__startup_cls is None
        assert self.Victim.result__teardown_cls is None

        # EXISTS IN INSTANCE --------------
        assert not hasattr(self.Victim, "INDEX")
        # assert not hasattr(self.Victim, "SETTINGS")
        assert hasattr(self.Victim, "DEV_COLUMN")

        assert not hasattr(self.Victim, "details")
        assert not hasattr(self.Victim, "exc")

    # -----------------------------------------------------------------------------------------------------------------
    def test__inst(self):
        self.Victim.STAND.DEV_LINES = Victim_DevicesLines()

        # EXISTS IN CLS --------------
        assert len(self.Victim.TCSi_LINE) == self.Victim.STAND.DEV_LINES.COUNT_COLUMNS

        # assert self.Victim(0) is self.Victim.TCSi_LINE[0]


        # TODO: FINISH!
        # TODO: FINISH!
        # TODO: FINISH!
        # TODO: FINISH!
        # TODO: FINISH!


# =====================================================================================================================



import random
import json
import string
import time


# Генерация случайного серийного номера
def generate_sn():
    letters = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
    digits = '0123456789'
    sn = random.choice(letters) + ''.join(random.choice(digits) for _ in range(5))
    return sn


# Генерация случайной записи о первичной производственной площадке
def generate_factory_record():
    factory = {
        "name": "Factory A",
        "stand_number": random.randint(1, 10),
        "shift_number": random.randint(1, 3)
    }
    return factory


# Генерация случайного результата теста
def generate_test_result():
    results = ["PASS", "FAIL"]
    return random.choice(results)


# Генерация случайной ссылки на базу компонент в составе изделия
def generate_components():
    components = {
        "component1": f"Component {random.choice(string.ascii_letters)}",
        "component2": f"Component {random.choice(string.ascii_letters)}"
    }
    return components


# Генерация случайных версий плат и прошивок в составе изделия
def generate_versions():
    versions = {
        "board_version": f"Version {random.randint(1, 5)}",
        "firmware_version": f"Version {random.randint(1, 10)}"
    }
    return versions


# Генерация случайного лога теста
def generate_test_log():
    log = {
        "test_name": "Test A",
        "test_result": generate_test_result(),
        "parameters": {
            "param1": random.randint(1, 100),
            "param2": random.randint(1, 100)
        },
        "log_file": "test_log.txt"
    }
    return log


# Генерация случайной записи истории изделия
def generate_history_record():
    history = {
        "status": random.choice(["Shipped", "Sent for repair", "Rejected", "In stock"]),
        "other_databases": {
            "database1": "Database A",
            "database2": "Database B"
        }
    }
    return history


def generate_data(num_records):
    data = []
    for _ in range(num_records):
        record = {
            "sn": generate_sn(),
            "timestamp": int(time.time()),
            "factory_record": generate_factory_record(),
            "test_result": generate_test_result(),
            "components": generate_components(),
            "versions": generate_versions(),
            "test_log": generate_test_log(),
            "history_record": generate_history_record()
        }
        data.append(record)
    return data


# ======================================================
def generate_data_1():
    record = {
        "sn": generate_sn(),
        "timestamp": int(time.time()),
        "factory_record": generate_factory_record(),
        "test_result": generate_test_result(),
        "components": generate_components(),
        "versions": generate_versions(),
        "test_log": generate_test_log(),
        "history_record": generate_history_record()
    }
    return record


# ======================================================
# for name, value in generate_data_1().items():
#     print(f"{name}: {value}")




from base_aux.base_lambdas.m1_lambda2_derivatives import *
from base_aux.aux_dict.m3_dict_ga import *


# =====================================================================================================================
dict_example = {
    "lowercase": "lowercase",
    # "nested": {"n1":1},
}


class Victim(DictIcKeys_Ga_AnnotRequired):
    lowercase: str


# =====================================================================================================================
def test__obj():
    # victim = DictIcKeys_Ga_AnnotRequired()
    # assert victim == {}
    #
    # victim = DictIcKeys_Ga_AnnotRequired(hello=1)
    # assert victim == {"hello": 1}

    try:
        victim = Victim()
    except:
        assert True
    else:
        assert False


def test__dict_only():
    assert Lambda_TrySuccess(DictIcKeys_Ga_AnnotRequired).resolve() == True
    assert Lambda_TrySuccess(DictIcKeys_Ga_AnnotRequired).resolve()

    assert Lambda_TryFail(DictIcKeys_Ga_AnnotRequired).resolve() != True
    assert not Lambda_TryFail(DictIcKeys_Ga_AnnotRequired).resolve()

    assert Lambda_TrySuccess(DictIcKeys_Ga_AnnotRequired, **dict_example).resolve()
    assert Lambda_TrySuccess(DictIcKeys_Ga_AnnotRequired, lowercase="lowercase").resolve()
    assert Lambda_TrySuccess(DictIcKeys_Ga_AnnotRequired, LOWERCASE="lowercase").resolve()


def test__with_annots():
    assert Lambda_TryFail(Victim).resolve()
    assert not Lambda_TrySuccess(Victim).resolve()

    victim = Victim(lowercase="lowercase")
    assert victim["lowercase"] == "lowercase"

    assert Lambda_TrySuccess(Victim, **dict_example).resolve()
    assert Lambda_TrySuccess(Victim, lowercase="lowercase").resolve()
    assert Lambda_TrySuccess(Victim, LOWERCASE="lowercase").resolve()

    assert Lambda_TryFail(Victim, hello="lowercase").resolve()

    victim = Victim(lowercase="lowercase")
    assert victim == {"lowercase": "lowercase"}

    # assert victim[1] == None
    assert Lambda_TryFail(lambda: victim[1]).resolve()

    assert victim.lowercase == "lowercase"


# =====================================================================================================================



import pytest

from base_aux.aux_attr.m4_kits import AttrKit_Blank
from base_aux.base_lambdas.m1_lambda import *
from base_aux.base_nest_dunders.m1_init2_annots1_attrs_by_args_kwargs import *
from base_aux.aux_eq.m3_eq_valid3_derivatives import *


# =====================================================================================================================
EQ_ISINSTANCE_VICTIM = EqValid_IsinstanceSameinstance(NestInit_AnnotsAttr_ByArgsKwargs)


class Test__NestInit:
    def test__notNested(self):
        try:
            AttrKit_Blank()
            # Init_AnnotsAttrsByKwArgsIc()
            NestInit_AnnotsAttr_ByArgsKwargs()
            # NestInit_AnnotsAttrByKwArgsIc()
        except:
            assert False

        assert AttrKit_Blank(a1=1).a1 == 1
        assert AttrKit_Blank(a1=1).A1 == 1

        assert NestInit_AnnotsAttr_ByArgsKwargs(a1=1).a1 == 1
        try:
            assert NestInit_AnnotsAttr_ByArgsKwargs(a1=1).A1 == 1
        except:
            pass
        else:
            assert False

        # assert Init_AnnotsAttrsByKwArgsIc(a1=1).a1 == 1
        # assert Init_AnnotsAttrsByKwArgsIc(a1=1).A1 == 1

    def test__Nested(self):
        class Example(NestInit_AnnotsAttr_ByArgsKwargs):
            A1: Any
            A2: Any = None
            A3 = None

        Example()

        # assert Example(a1=1).A1 == 1
        # assert Example(1, a1=2).A1 == 2

        assert Example(1).A1 == 1
        assert Example(1).A2 == None
        assert Example(1).A3 == None

        try:
            assert Example(1, 1, 1)
        except:
            pass

        assert Example(1, 1).A2 == 1
        assert Example(1, 1).A3 == None
        assert Example(1, 1, a3=1).A3 == 1

    # -----------------------------------------------------------------------------------------------------------------
    @pytest.mark.parametrize(
        argnames="args, kwargs, _EXPECTED",
        argvalues=[
            ((), dict(k1=1), EQ_ISINSTANCE_VICTIM),
            ((), dict(k1=1, k2=2), EQ_ISINSTANCE_VICTIM),

            ((), dict(k1=1, k2=2), EQ_ISINSTANCE_VICTIM),
        ]
    )
    def test__1(self, args, kwargs, _EXPECTED):
        Lambda(NestInit_AnnotsAttr_ByArgsKwargs, *args, **kwargs).check_expected__assert(_EXPECTED)

        if _EXPECTED == Exception:
            return

        victim = NestInit_AnnotsAttr_ByArgsKwargs(*args, **kwargs)
        for key, value in kwargs.items():
            assert getattr(victim, key) == value

        for arg in args:
            # args used only for Annots!
            try:
                getattr(victim, arg)
            except:
                pass
            else:
                assert False

# =====================================================================================================================
class Victim(NestInit_AnnotsAttr_ByArgsKwargs):
    # At0
    At1 = None
    An0: Any
    An1: Any = None


# ---------------------------------------------------------------------------------------------------------------------
@pytest.mark.parametrize(
    argnames="args, kwargs, _EXPECTED, values",
    argvalues=[
        ((), dict(At0=111), Victim, (111, None, Exception, None)),
        ((), dict(At1=222), Victim, (Exception, 222, Exception, None)),

        ((333, 444), dict(At0=111, At1=222), EQ_ISINSTANCE_VICTIM, (111, 222, 333, 444)),
        ((333, 444), dict(At0=111, At1=222), EQ_ISINSTANCE_VICTIM, (111, 222, 333, 444)),
        ((11, 22), dict(At0=111, At1=222, An0=333, An1=444), EQ_ISINSTANCE_VICTIM, (111, 222, 333, 444)),
        ((11, 22), dict(AT0=111, AT1=222, AN0=333, AN1=444), EQ_ISINSTANCE_VICTIM, (Exception, 222, 333, 444)),
    ]
)
def test__2(args, kwargs, _EXPECTED, values):
    Lambda(Victim, *args, **kwargs).check_expected__assert(_EXPECTED)

    victim = Victim(*args, **kwargs)
    for index, name in enumerate(["At0", "At1", "An0", "An1"]):
        Lambda(getattr, victim, name).check_expected__assert(values[index])


# =====================================================================================================================



from typing import *
from base_aux.base_values.m2_value_special import NoValue

from base_aux.aux_attr.m1_annot_attr1_aux import AttrAux_Existed


# =====================================================================================================================
class Translator:
    """
    SPECIALLY CREATED FOR
    ---------------------
    translate breeder_stack values into russian
    """
    RETURN_SOURCE_IF_NOT_FOUND: bool = True

    SELECTOR: Callable[[Any, Any], bool] = lambda self, source, var: source == var
    RULES: dict[Any | str, Any]

    def __init__(self, rules: dict[Any, Any] | Any, selector=None, return_source_if_not_found: bool = None):
        if selector is not None:
            self.SELECTOR = selector
        if rules is not None:
            self.RULES = rules
        if return_source_if_not_found is not None:
            self.RETURN_SOURCE_IF_NOT_FOUND = return_source_if_not_found

        if not isinstance(self.RULES, dict):
            self.RULES = AttrAux_Existed(self.RULES).dump_dict()

    def __call__(self, source: Any, *args, **kwargs) -> Any | type[NoValue]:
        """
        GOAL
        ----
        actually make translation

        any raise - assumed as not valid/selected
        """
        for variant, result in self.RULES.items():
            try:
                if self.SELECTOR(source, variant):
                    return result
            except:
                pass

        if self.RETURN_SOURCE_IF_NOT_FOUND:
            return source
        else:
            return NoValue


# =====================================================================================================================
if __name__ == "__main__":
    victim = Translator({1:11, 2:22})
    assert victim(1) == 11
    assert victim(2) == 22
    assert victim(3) == 3

    victim = Translator({1:11, 2:22}, return_source_if_not_found=False)
    assert victim(3) == NoValue
    # run_over_api()


# =====================================================================================================================



from typing import *

from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *

from base_aux.pyqt.m1_dialog import Dialogs


# SET ============================================================================================================
class DialogsTp(Dialogs):
    @classmethod
    def info__about(cls, *args) -> int:
        wgt = QMessageBox()
        if not cls._apply_new__or_activate_last(name="info__about", wgt=wgt):
            return 1024
        wgt.setMaximumWidth(1000)
        answer = wgt.information(
            None,
            "О программе",
            (
                "LLC,\n"
                "Программа проведения тестирования"
             )
        )
        # return always 1024
        return answer

    @classmethod
    def finished__devs_detection(cls, *args) -> int:
        wgt = QMessageBox()
        if not cls._apply_new__or_activate_last(name="finished__devs_detection", wgt=wgt):
            return 1024
        wgt.resize(1000, 1000)
        wgt.setBaseSize(1000, 1000)
        answer = wgt.information(
            None,
            "Определение устройств",
            (
                "Процесс завершен" + " "*30
            )
        )
        # return always 1024
        return answer

    @classmethod
    def finished__tp(cls, *args) -> int:
        wgt = QMessageBox()
        if not cls._apply_new__or_activate_last(name="finished__tp", wgt=wgt):
            return 1024
        answer = wgt.information(
            None,
            "Тестирование",
            (
                "Процесс завершен"
             )
        )
        # return always 1024
        return answer

    @classmethod
    def finished__save(cls, *args) -> int:
        wgt = QMessageBox()
        if not cls._apply_new__or_activate_last(name="finished__save", wgt=wgt):
            return 1024
        answer = wgt.information(
            None,
            "Сохранение",
            (
                "Процесс завершен"
             )
        )
        # return always 1024
        return answer


# =====================================================================================================================
if __name__ == '__main__':
    # DialogsTp.info__about()
    DialogsTp.finished__devs_detection()
    # DialogsTp.finished__tp()
    # DialogsTp.finished__save()


# =====================================================================================================================



import pytest

from base_aux.base_lambdas.m1_lambda import *
from base_aux.base_nest_dunders.m4_gsai_annots import NestGAI_AnnotAttrIC
from base_aux.aux_attr.m4_kits import *
from base_aux.aux_eq.m3_eq_valid3_derivatives import *


# =====================================================================================================================
class VictimAt:
    at1 = 1
    _at1 = 11
    __at1 = 111


class VictimAn:
    an1: int = 2
    _an1: int = 22
    __an1: int = 222


class VictimAnNest(VictimAn):
    an2: int = 3
    _an2: int = 33
    __an2: int = 333


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="source, skip_names, _EXPECTED",
    argvalues=[
        (VictimAt(), [], [
            {"at1", "_at1", "__at1"},
            {"at1", "_at1", "__at1"},
            {"at1", "_at1", "__at1"},
        ]),
        (VictimAn(), [], [
            {"an1", "_an1", "__an1"},
            {"an1", "_an1", "__an1"},
            {"an1", "_an1", "__an1"},
        ]),
        (VictimAnNest(), [], [
            {"an1", "_an1", "__an1", "an2", "_an2", "__an2"},
            {"an1", "_an1", "__an1", "an2", "_an2", "__an2"},
            {"an1", "_an1", "__an1", "an2", "_an2", "__an2"},
        ]),

        # -------------------------------
        (VictimAt(), ["at1", ], [
            {"_at1", "__at1"},
            {"_at1", "__at1"},
            {"_at1", "__at1"},
        ]),
        (VictimAn(), ["an1", ], [
            {"_an1", "__an1"},
            {"_an1", "__an1"},
            {"_an1", "__an1"},
        ]),
        (VictimAnNest(), ["an1", ], [
            {"_an1", "__an1", "an2", "_an2", "__an2"},
            {"_an1", "__an1", "an2", "_an2", "__an2"},
            {"_an1", "__an1", "an2", "_an2", "__an2"},
        ]),

        # -------------------------------
        (VictimAt(), [EqValid_ContainStrIc("_a"), ], [
            {"at1", },
            {"at1", },
            {"at1", },
        ]),
        (VictimAn(), [EqValid_ContainStrIc("_a"), ], [
            {"an1", },
            {"an1", },
            {"an1", },
        ]),
        (VictimAnNest(), [EqValid_ContainStrIc("_a"), ], [
            {"an1", "an2", },
            {"an1", "an2", },
            {"an1", "an2", },
        ]),
    ]
)
def test__ITER_1__iter__dirnames_original_not_builtin(source, skip_names, _EXPECTED):
    Lambda(set(AttrAux_Existed(source, *skip_names).iter__dirnames_original_not_builtin())).check_expected__assert(set(_EXPECTED[0]))
    Lambda(set(AttrAux_AnnotsAll(source, *skip_names).iter__dirnames_original_not_builtin())).check_expected__assert(set(_EXPECTED[1]))
    Lambda(set(AttrAux_AnnotsLast(source, *skip_names).iter__dirnames_original_not_builtin())).check_expected__assert(set(_EXPECTED[2]))


# =====================================================================================================================
# NOTE: add iter_names - NOT NEED! used as next exact methods


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="source, skip_names, _EXPECTED",
    argvalues=[
        (VictimAt(), [], [
            {"at1", },
            {},
            {},
        ]),
        (VictimAn(), [], [
            {"an1", },
            {"an1", },
            {"an1", },
        ]),
        (VictimAnNest(), [], [
            {"an1", "an2", },
            {"an1", "an2", },
            {"an2", },
        ]),
    ]
)
def test__ITER__iter__names_filter__not_hidden(source, skip_names, _EXPECTED):
    Lambda(set(AttrAux_Existed(source, *skip_names).iter__names_filter__not_hidden())).check_expected__assert(set(_EXPECTED[0]))
    Lambda(set(AttrAux_AnnotsAll(source, *skip_names).iter__names_filter__not_hidden())).check_expected__assert(set(_EXPECTED[1]))
    Lambda(set(AttrAux_AnnotsLast(source, *skip_names).iter__names_filter__not_hidden())).check_expected__assert(set(_EXPECTED[2]))


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="source, skip_names, _EXPECTED",
    argvalues=[
        (VictimAt(), [], [
            {"at1", "_at1", },
            {},
            {},
        ]),
        (VictimAn(), [], [
            {"an1", "_an1", },
            {"an1", "_an1", },
            {"an1", "_an1", },
        ]),
        (VictimAnNest(), [], [
            {"an1", "an2", "_an1", "_an2", },
            {"an1", "an2", "_an1", "_an2", },
            {"an2", "_an2", },
        ]),
    ]
)
def test__ITER__iter__names_filter__not_private(source, skip_names, _EXPECTED):
    Lambda(set(AttrAux_Existed(source, *skip_names).iter__names_filter__not_private())).check_expected__assert(set(_EXPECTED[0]))
    Lambda(set(AttrAux_AnnotsAll(source, *skip_names).iter__names_filter__not_private())).check_expected__assert(set(_EXPECTED[1]))
    Lambda(set(AttrAux_AnnotsLast(source, *skip_names).iter__names_filter__not_private())).check_expected__assert(set(_EXPECTED[2]))


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="source, skip_names, _EXPECTED",
    argvalues=[
        (VictimAt(), [], [
            {"__at1", },
            {},
            {},
        ]),
        (VictimAn(), [], [
            {"__an1", },
            {"__an1", },
            {"__an1", },
        ]),
        (VictimAnNest(), [], [
            {"__an1", "__an2", },
            {"__an1", "__an2", },
            {"__an2", },
        ]),
    ]
)
def test__ITER__iter__names_filter__private(source, skip_names, _EXPECTED):
    Lambda(set(AttrAux_Existed(source, *skip_names).iter__names_filter__private())).check_expected__assert(set(_EXPECTED[0]))
    Lambda(set(AttrAux_AnnotsAll(source, *skip_names).iter__names_filter__private())).check_expected__assert(set(_EXPECTED[1]))
    Lambda(set(AttrAux_AnnotsLast(source, *skip_names).iter__names_filter__private())).check_expected__assert(set(_EXPECTED[2]))


# =====================================================================================================================
# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="source, skip_names, _EXPECTED",
    argvalues=[
        (VictimAt(), [], [
            {VictimAt, },
            {VictimAt, },
            {VictimAt, },
        ]),
        (VictimAn(), [], [
            {VictimAn, },
            {VictimAn, },
            {VictimAn, },
        ]),
        (VictimAnNest(), [], [
            {VictimAnNest, VictimAn},
            {VictimAnNest, VictimAn},
            {VictimAnNest, VictimAn},
        ]),
    ]
)
def test__ITER___iter_mro(source, skip_names, _EXPECTED):
    Lambda(set(AttrAux_Existed(source, *skip_names)._iter_mro())).check_expected__assert(set(_EXPECTED[0]))
    Lambda(set(AttrAux_AnnotsAll(source, *skip_names)._iter_mro())).check_expected__assert(set(_EXPECTED[1]))
    Lambda(set(AttrAux_AnnotsLast(source, *skip_names)._iter_mro())).check_expected__assert(set(_EXPECTED[2]))


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="source, skip_names, _EXPECTED",
    argvalues=[
        (VictimAt(), [], [
            {},
            {},
            {},
        ]),
        (VictimAn(), [], [
            {"an1", "_an1", "__an1"},
            {"an1", "_an1", "__an1"},
            {"an1", "_an1", "__an1"},
        ]),
        (VictimAnNest(), [], [
            {"an1", "_an1", "__an1", "an2", "_an2", "__an2"},
            {"an1", "_an1", "__an1", "an2", "_an2", "__an2"},
            {"an2", "_an2", "__an2"},
        ]),
    ]
)
def test__ITER__iter__annot_names(source, skip_names, _EXPECTED):
    Lambda(set(AttrAux_Existed(source, *skip_names).iter__annot_names())).check_expected__assert(set(_EXPECTED[0]))
    Lambda(set(AttrAux_AnnotsAll(source, *skip_names).iter__annot_names())).check_expected__assert(set(_EXPECTED[1]))
    Lambda(set(AttrAux_AnnotsLast(source, *skip_names).iter__annot_names())).check_expected__assert(set(_EXPECTED[2]))


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="source, skip_names, _EXPECTED",
    argvalues=[
        (VictimAt(), [], [
            {},
            {},
            {},
        ]),
        (VictimAn(), [], [
            {2, 22, },
            {2, 22, },
            {2, 22, },
        ]),
        (VictimAnNest(), [], [
            {2, 22, 3, 33},
            {2, 22, 3, 33},
            {3, 33, },
        ]),
    ]
)
def test__ITER__iter__annot_values(source, skip_names, _EXPECTED):
    Lambda(set(AttrAux_Existed(source, *skip_names).iter__annot_values())).check_expected__assert(set(_EXPECTED[0]))
    Lambda(set(AttrAux_AnnotsAll(source, *skip_names).iter__annot_values())).check_expected__assert(set(_EXPECTED[1]))
    Lambda(set(AttrAux_AnnotsLast(source, *skip_names).iter__annot_values())).check_expected__assert(set(_EXPECTED[2]))


# =====================================================================================================================



from typing import *
from pydantic import BaseModel


# =====================================================================================================================
TYPES__DICT = dict[str, Union[None, str, bool, int, float, dict, list]]


# =====================================================================================================================
class Model_StandInfo(BaseModel):
    # STAND_NAME: str           # "StandPSU"
    # STAND_DESCRIPTION: str    # "test PSU for QCD"
    # STAND_SN: str
    STAND_SETTINGS: TYPES__DICT = {}     # main settings for all TCS


class Model_DeviceInfo(BaseModel):
    INDEX: int          # device position in stand

    NAME: str           # "PSU"
    DESCRIPTION: str    # "Power Supply Unit"
    SN: str


class Model_TcInfo(BaseModel):
    TC_NAME: str
    TC_DESCRIPTION: str

    TC_ASYNC: bool
    TC_SKIP: bool

    TC_SETTINGS: TYPES__DICT = {
        # CONTENT IS NOT SPECIFIED!
        # "ANY_1": Any,
    }


class Model_TcResult(BaseModel):
    tc_timestamp: float | None = None

    tc_active: bool = False
    tc_result: bool | None = None
    tc_details: TYPES__DICT = {
        # CONTENT IS NOT SPECIFIED!
        # "ANY_2": Any,
    }


# =====================================================================================================================
class Model_TcResultFull(Model_TcResult, Model_TcInfo, Model_DeviceInfo):
    pass


class Model_TpInfo(Model_StandInfo):
    TESTCASES: list[Model_TcInfo]


class Model_TpResults(Model_StandInfo):
    TESTCASES: list[list[Model_TcResultFull]]


# =====================================================================================================================



from typing import *

from base_aux.aux_np_pd.m3_indicators import *


# =====================================================================================================================
class Indicator_Wma(Base_Indicator):
    """
    COLUMN_NAME__TEMPLATE = WMA_%(length)s
    """
    NAME = "WMA"
    COLUMN_SETINGS = dict(
        WMA=ColumnSettings(EqValid_Regexp(r"WMA_\d+"), 1),
    )
    PARAMS: DictIc_LockedKeys_Ga = DictIc_LockedKeys_Ga(
        length=12,
    )

    # results -----
    WMA: Any

    @property
    def TA_METH(self) -> Callable[..., Any]:
        return self.DF.ta.wma

    @property
    def HISTORY_ENOUGH_THRESH(self) -> int:
        return self.PARAMS.length


# ---------------------------------------------------------------------------------------------------------------------
class Indicator_Rsi(Base_Indicator):
    """
    length: int

    COLUMN_NAME__TEMPLATE: str = "RSI_%(length)s"
    ROUND: int = 1
    """
    NAME = "RSI"
    COLUMN_SETINGS = dict(
        RSI=ColumnSettings(EqValid_Regexp(r"RSI_\d+"), 1),
    )
    PARAMS: DictIc_LockedKeys_Ga = DictIc_LockedKeys_Ga(
        length=12,
    )

    # results -----
    RSI: Any

    @property
    def TA_METH(self) -> Callable[..., Any]:
        return self.DF.ta.rsi

    @property
    def HISTORY_ENOUGH_THRESH(self) -> int:
        return self.PARAMS.length


# =====================================================================================================================
class Indicator_Adx(Base_Indicator):
    """
    length: int
    lensig: int

    "ADX_%(lensig)s"
    """
    NAME = "ADX"
    COLUMN_SETINGS = dict(
        ADX=ColumnSettings(EqValid_Regexp(r"ADX_\d+"), 1),
        DMP=ColumnSettings(EqValid_Regexp(r"DMP_\d+"), 1),
        DMN=ColumnSettings(EqValid_Regexp(r"DMN_\d+"), 1),
    )
    PARAMS: DictIc_LockedKeys_Ga = DictIc_LockedKeys_Ga(
        length=13,
        lensig=9,
    )

    # results -----
    ADX: Any
    DMP: Any
    DMN: Any

    @property
    def TA_METH(self) -> Callable[..., Any]:
        return self.DF.ta.adx

    @property
    def HISTORY_ENOUGH_THRESH(self) -> int:
        return sum(self.PARAMS.values()) * 10


# ---------------------------------------------------------------------------------------------------------------------
class Indicator_Macd(Base_Indicator):
    """
    fast: int
    slow: int
    signal: int

    ROUND: int = 3

    @property
    def COLUMN_NAME__TEMPLATE(self) -> str:
        if self.slow < self.fast:
            return "MACDh_%(slow)s_%(fast)s_%(signal)s"
        else:
            return "MACDh_%(fast)s_%(slow)s_%(signal)s"
    """
    NAME = "MACD"
    COLUMN_SETINGS = dict(
        MACD=ColumnSettings(EqValid_Regexp(r"MACD(?:_\d+){3}"), 3),     # check
        HIST=ColumnSettings(EqValid_Regexp(r"MACDh(?:_\d+){3}"), 3),
        SIG=ColumnSettings(EqValid_Regexp(r"MACDs(?:_\d+){3}"), 3),  # check
    )
    PARAMS: DictIc_LockedKeys_Ga = DictIc_LockedKeys_Ga(
        fast=12,
        slow=26,
        signal=9,
    )

    # results -----
    MACD: Any
    HIST: Any
    SIG: Any

    @property
    def TA_METH(self) -> Callable[..., Any]:
        return self.DF.ta.macd

    @property
    def HISTORY_ENOUGH_THRESH(self) -> int:
        return sum(self.PARAMS.values()) * 10


# ---------------------------------------------------------------------------------------------------------------------
class Indicator_Stoch(Base_Indicator):
    """
    always work with 14/3/3!!!

    fast_k: int
    slow_k: int
    slow_d: int

    COLUMN_NAME__TEMPLATE: str = "STOCHk_%(fast_k)s_%(slow_k)s_%(slow_d)s"
    COLUMN_NAME__TEMPLATE: str = "STOCHk_14_3_3"
    """
    NAME = "STOCH"
    COLUMN_SETINGS = dict(
        STOCH=ColumnSettings(EqValid_Regexp(r"STOCHk(?:_\d+){3}"), 1),
        STOCHd=ColumnSettings(EqValid_Regexp(r"STOCHd(?:_\d+){3}"), 1),
    )
    PARAMS: DictIc_LockedKeys_Ga = DictIc_LockedKeys_Ga(
        fast_k=3,
        slow_k=14,
        slow_d=3,
    )

    # results -----
    STOCH: Any
    STOCHd: Any

    @property
    def TA_METH(self) -> Callable[..., Any]:
        return self.DF.ta.stoch

    @property
    def HISTORY_ENOUGH_THRESH(self) -> int:
        return self.PARAMS.slow_k


# =====================================================================================================================



from base_aux.base_values.m3_exceptions import *

from base_aux.aux_attr.m2_annot3_gacls_keys_as_values import NestGaCls_AnnotNamesAsValuesIc
from base_aux.numbers.m1_arithm import ArithmApplyToAttr

from base_aux.aux_eq.m2_eq_aux import *


# =====================================================================================================================
class UnitBase(NestGaCls_AnnotNamesAsValuesIc):
    # class UnitBase:
    """
    this is just an example! it will help you to use just strings by Object
    """
    V: str
    A: str
    C: str


UNIT_MULT__VARIANTS: dict[str, float | int] = {
    # ENG/INTERNATIONAL
    "p": 10 ** (-12),
    "n": 10 ** (-9),
    "µ": 10 ** (-6),
    "m": 10 ** (-3),
    "c": 10 ** (-2),
    "d": 10 ** (-1),

    "k": 10 ** 3,
    "M": 10 ** 6,
    "G": 10 ** 9,
    "T": 10 ** 12,

    # RUS
    "п": 10 ** (-12),
    "н": 10 ** (-9),
    "мк": 10 ** (-6),
    "м": 10 ** (-3),
    "с": 10 ** (-2),
    "д": 10 ** (-1),

    "к": 10 ** 3,
    "М": 10 ** 6,
    "Г": 10 ** 9,
    "Т": 10 ** 12,

    # ONE
    "": 10 ** 0,  # keep it last only or just delete!!!
}


# =====================================================================================================================
pass


# TODO: rename to ValueUnit_Arithm
class ValueUnit(ArithmApplyToAttr):
    """
    GOAL
    ----
    keep separated/parse VALUE and measure UNIT and compare with any representation

    CREATED SPECIALLY FOR
    ---------------------
    use in UART/SERIAL to get exact inline comparable numeric value!

    NOTE
    ----
    DONT MESS VALUE&VALUE_PURE!!!
        VALUE - is an original number with multiplier not applied
        VALUE_PURE - multiplier applied

    Example:
        SOURCE = 1k
        VALUE = 1
        VALUE_PURE = 1000

    so VALUE - is the most useful representation in your application!!!

    BEST USAGE
    ----------
        assert ValueUnit('0.0k') == 0
        assert ValueUnit('0,01k') == 10

        assert ValueUnit('1нм') == '1n'    #RUS multipliers are acceptable! and mean the same!

        assert ValueUnit('1k') == '1000'
        assert ValueUnit('1kV') == 1000
        assert ValueUnit('1kV') == '1k'
        assert ValueUnit(1000) == '1kV'

        assert ValueUnit(1001) > '1kV'
        assert ValueUnit(1000) < '1.1kV'

        - NoValue - same as object as pattern!

        assert "1" == ValueUnit(NoValue)
        assert "1V" == ValueUnit(NoValue)
        assert "1V" == ValueUnit(NoValue, unit="V")
        assert "1" == ValueUnit(NoValue, unit="V")
        assert "1A" != ValueUnit(NoValue, unit="V")

    """
    # NESTED ----------------------
    NUMBER_ARITHM__ATTR = "VALUE_PURE"

    # SETTINGS -------------------------
    SEPARATOR: str = ""     # when parsed - used pared separator, when

    UNIT_MULT__DISABLE: bool | None = None
    UNIT_MULT__VARIANTS: dict[str, float | int] = UNIT_MULT__VARIANTS

    # NEW -------------------------
    UNIT: str = ""      # final unit with multiplier
    UNIT_MULT: str = ""
    UNIT_BASE: str = ""
    MULT: int = 1

    SOURCE: Any = None
    VALUE: Union[int, float] = 0    # DONT MESS VALUE/VALUE_PURE!!! see NOTES! if UNIT_MULT__DISABLE all are the same!!!

    @property
    def VALUE_PURE(self) -> Union[int, float]:
        return self.number__fix_precision(self.VALUE * self.MULT)

    @VALUE_PURE.setter
    def VALUE_PURE(self, value_pure: Any) -> None:
        value_pure = self._other__get_float(value_pure)
        if int(value_pure) == float(value_pure):
            value_pure = int(value_pure)

        value = value_pure / self.MULT
        value = self.number__fix_precision(value)
        self.VALUE = value

    # -----------------------------------------------------------------------------------------------------------------
    def __init__(self, source: Union[int, float, str, Any] = NoValue, unit: str = None, separator: str = None, mult_disable: bool = None):
        """
        :param source:
        :param unit: use it only if not exists in source!
        :param separator: if set - would overwrite parsed!
        """
        if mult_disable is not None:
            self.UNIT_MULT__DISABLE = mult_disable

        if self.UNIT_MULT__DISABLE:
            self.UNIT_MULT__VARIANTS = {}

        self.SOURCE = source

        # first parse -----------------
        if source is NoValue:
            pass
        else:
            self.parse(source)

        # second OVERWRITE  -----------------
        if unit is not None:
            self.UNIT = unit
            self.UNIT_BASE = unit
            self.UNIT_MULT = ""
            self.MULT = 1
        if separator is not None:
            self.SEPARATOR = separator

    @classmethod
    def validate(cls, source: Any) -> bool:
        """
        CREATED SPECIALLY FOR
        ---------------------
        if ypu want to decide somewhere return ValueUnit-object or just source
        check source before applying
        """
        try:
            cls(source)
            return True
        except:
            return False

    def clear(self) -> None:
        self.SOURCE = None
        self.VALUE = 0
        self.UNIT = ""
        self.UNIT_MULT = ""
        self.UNIT_BASE = ""
        self.MULT = 1

    def parse(self, source: Any) -> Self | NoReturn:
        self.clear()
        self.SOURCE = source

        if isinstance(source, (int, float)):
            self.VALUE = source
        else:
            source = str(source)
            source = source.strip()
            source = source.replace(',', ".")
            source = re.sub(r'-\s*', '-', source)
            source = re.sub(r'\+\s*', '', source)
            source = re.sub(r'\'\'', '', source)
            source = re.sub(r'\"\"', '', source)

            # WORK ---------------------------
            match = re.fullmatch(r'(-?[0-9.]+)(\s*)([a-zA-Zа-яА-Я]*)', source)
            if match:
                self.VALUE = float(match[1])
                self.SEPARATOR = match[2] or ""
                self.UNIT = match[3] or ""
            else:
                raise Exc__NotExistsNotFoundNotCreated()

            # UNIT_MULT ------------------
            if not self.UNIT_MULT__DISABLE:
                for unit_mult, multiplier in self.UNIT_MULT__VARIANTS.items():
                    if self.UNIT.startswith(unit_mult):
                        self.UNIT_MULT = unit_mult
                        self.UNIT_BASE = self.UNIT.removeprefix(unit_mult)
                        self.MULT = multiplier
                        break

        # PRECISION -----------------------
        self.VALUE = self.number__fix_precision(self.VALUE)

        # INT -----------------------------
        if int(self.VALUE) == self.VALUE:
            self.VALUE = int(self.VALUE)
        else:
            self.VALUE = self.VALUE

        # if not self.UNIT_BASE and self.UNIT:
        #     self.UNIT_BASE = self.UNIT

        # FINISH ---------------------------
        return self

    def __str__(self) -> str:
        return f"{self.VALUE}{self.SEPARATOR}{self.UNIT}"

    def __repr__(self) -> str:
        """
        used as help
        """
        return str(self)

    # CMP -------------------------------------------------------------------------------------------------------------
    def __cmp__(self, other) -> int | NoReturn:
        """
        do try to resolve Exceptions!!! sometimes it is ok to get it!!!

        RETURN
        ------
            1=self>other
            0=self==other
            -1=self<other
        """
        if EqAux(self.VALUE).check_doubleside__bool(other):
            return 0
        if not isinstance(other, self.__class__):
            other = self.__class__(other)

        if self.SOURCE == NoValue or other.SOURCE == NoValue:
            if self.UNIT_BASE == other.UNIT_BASE or self.UNIT == other.UNIT:
                return 0
            elif not self.UNIT and not other.UNIT:
                return 0
            else:
                return -1   # just to show not equel!!!

        # cmp units -----------
        if self.UNIT_BASE and other.UNIT_BASE and self.UNIT_BASE != other.UNIT_BASE:
            msg = f"{self!r}/{other!r}"
            raise Exc__Incompatible_Data(msg)

        # cmp -----------
        if self.VALUE_PURE == other.VALUE_PURE:
            return 0
        elif self.VALUE_PURE > other.VALUE_PURE:
            return 1
        elif self.VALUE_PURE < other.VALUE_PURE:
            return -1


# =====================================================================================================================
class ValueUnit_NoMulty(ValueUnit):
    """
    this is just as parser! when you need only get number near the unit without object arithmetic acceptable.
    but with comparing methods.

    NOTE
    ----
    maybe it is cant be used
    """
    UNIT_MULT__DISABLE = True


# =====================================================================================================================



import pytest


class Cls:
    A1 = 1
    A2 = 2


class Test_1:
    CLS = Cls

    @pytest.mark.parametrize(
        argnames="attr, _EXPECTED",
        argvalues=[
            (CLS.A1, 1),
            (CLS.A2, 2),
        ]
    )
    def test__1(self, attr, _EXPECTED):
        assert attr == _EXPECTED


class Cls2:
    A1 = 11
    A2 = 22


class Test_2:
    CLS = Cls2




from typing import *
import pytest
import asyncio
from pytest import mark

from base_aux.base_lambdas.m1_lambda import *


# =====================================================================================================================
@pytest.mark.asyncio
@pytest.mark.parametrize(
    argnames="func_link, args, _EXPECTED",
    argvalues=[
        (asyncio.sleep, (0.1, 1), 1),
        (asyncio.sleep, ("hello",), TypeError),
    ]
)
async def test__aio(func_link, args, _EXPECTED):
    try:
        result = await func_link(*args)
    except BaseException as exc:
        result = type(exc)

    assert result == _EXPECTED

"""
Testing started at 18:19 ...
Launching pytest with arguments test0_tasks1_cm.py::test__aio --no-header --no-summary -q in C:\__STARICHENKO_Element\PROJECTS\abc=aio_testplan\aio_tp2

============================= test session starts =============================
collecting ... collected 2 items

test0_tasks1_cm.py::test__aio[sleep-args0-1] 
test0_tasks1_cm.py::test__aio[sleep-args1-TypeError] 

============================== 2 passed in 0.16s ==============================
PASSED                      [ 50%]PASSED              [100%]sys:1: RuntimeWarning: coroutine 'sleep' was never awaited

Process finished with exit code 0
"""


# =====================================================================================================================



from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart

import smtplib

from base_aux.alerts.m1_alert0_base import *
from base_aux.privates.m1_privates import *
from base_aux.aux_attr.m4_kits import *


# =====================================================================================================================
class SmtpServers:
    """well known servers addresses.

    Here we must collect servers like MilRu/GmailCom, and not to create it in any new project.
    """
    MAIL_RU: AttrKit_AddrPort = AttrKit_AddrPort("smtp.mail.ru", 465)


# =====================================================================================================================
class AlertSmtp(Base_Alert):
    """
    GOAL
    ----
    SMTP realisation for sending msg (email).

    :param _subtype: reuse new _subtype instead of default
    """
    # SETTINGS ------------------------------------
    CONN_ADDRESS: AttrKit_AddrPort = SmtpServers.MAIL_RU
    CONN_AUTH: AttrKit_AuthNamePwd = PvLoaderIni_AuthNamePwd(keypath=("AUTH_EMAIL_DEF",))
    TIMEOUT_SEND = 5

    # AUX -----------------------------------------
    _conn:  smtplib.SMTP_SSL
    _subtype: str = "plain"
    SUBJECT: str = None

    def __init__(self, *args, _subtype: str = None, subject: str = None, **kwargs):
        if _subtype is not None:
            self._subtype = _subtype
        if subject is not None:
            self.SUBJECT = subject

        super().__init__(*args, **kwargs)

    def _connect_unsafe(self) -> Union[bool, NoReturn]:
        self._conn = smtplib.SMTP_SSL(self.CONN_ADDRESS.ADDR, self.CONN_ADDRESS.PORT, timeout=self.TIMEOUT_SEND)
        return True

    def _login_unsafe(self) -> Union[bool, NoReturn]:
        response = self._conn.login(self.CONN_AUTH.NAME, self.CONN_AUTH.PWD)
        print(response)
        print("=" * 100)
        return response and response[0] in [235, 503]

    def _send_unsafe(self) -> Union[bool, NoReturn]:
        self._conn.send_message(self._msg_compose())
        return True

    def _msg_compose(self) -> MIMEMultipart:
        msg = MIMEMultipart()
        msg["From"] = self.CONN_AUTH.NAME
        msg["To"] = self.RECIPIENT
        msg['Subject'] = self.SUBJECT or self.__class__.__name__

        try:
            _subtype = self.MSG_ACTIVE["_subtype"]
        except:
            _subtype = self._subtype
        msg.attach(MIMEText(self.MSG_ACTIVE, _subtype=_subtype))
        return msg

    def _recipient_get(self) -> str:
        return self.CONN_AUTH.NAME


# =====================================================================================================================
if __name__ == "__main__":
    victim = AlertSmtp()
    victim.send_msg("hello")
    victim.send_msg("hello2")
    victim.wait()
    victim.send_msg("hello3")
    victim.wait()


# =====================================================================================================================



from base_aux.base_types.m0_static_typing import *
from base_aux.base_lambdas.m1_lambda import *


# =====================================================================================================================
class Base_EqValid(NestCall_Resolve):
    """
    RULES
    -----
    1/ ALL not defined *ARGS/**KWARGS WHAT PASSED INTO INIT WOULD PASS INTO VALIDATOR() after other_final

    NOTE
    ----
    1/ preferably not use directly this object!
    USE DERIVATIVES!!! without validator passing

    2/ MAIN IDEA - NEVER RAISED!!! if any - return FALSE!!! if need - check manually!
    why so? - because i need smth to make a tests with final result of any source!
    dont mind reason!

    GOAL
    ----
    base object to make a validation by direct comparing with other object
    no raise

    USAGE
    -----
    for testing some other value with EqValidator
    1/ create validator object with exact params
    2/ use next operators for validating
        - EQ(eqObj == otherValue)
        - CONTAIN(otherValue in eqObj)
    No matter what eqObj is doing CONTAIN always will work!

    IF raised on any variant - pass to nest variant!

    USING TRICK (GOOD PRACTICE)
    ---------------------------
    any parameter could have exclude/skip variants with cmp by direct EQ operator for each value or IN operator for all values
    you can change logic by ising EqValid values instead of simple generics
    Example
        def dump_attrs(obj: Any, skip_names: list[str] = None):
            pass
            ...

        dump_attrs(Cls(), skip_names=["exit", call1, call2, call3])
        dump_attrs(Cls(), skip_names=["exit", EqValid_Startswith(call)])

    IRESULT_REVERSE
    ===============
    it mean resolve direct result anв only after that reverse exact final value!!!

    IRESULT_REVERSE vs IRESULT_CUMULATE
    -----------------------------------
    NOTE: use only one param! - IRESULT_REVERSE or IRESULT_CUMULATE!!!
      IRESULT_REVERSE - used for validation result!
      IRESULT_CUMULATE - used for cumulating ARGS validation results! - ignored FOR SINGLE result!
      DONT add REVERSE FinalResult! it intended in IRESULT_CUMULATE!!!

    """
    VALIDATOR: TYPING.VALID_VALIDATOR    # DEFINE!!!

    V_ARGS: TYPING.ARGS_FINAL       # as variant for validation! can be blank!
    V_KWARGS: TYPING.KWARGS_FINAL   # as settings!

    # IRESULT_REVERSE vs IRESULT_CUMULATE: SEE notes in docstr!
    IRESULT_REVERSE: bool = None
    IRESULT_CUMULATE: EnumAdj_BoolCumulate = EnumAdj_BoolCumulate.ALL_TRUE

    OTHER_FINAL__RESOLVE: bool = True   # goal: create chains with no ReCalculation inside
    OTHER_RAISED: bool = None   # RAISE ON CALCULATION OTHER_FINAL! not on equation!!!
    OTHER_FINAL: Any | Exception = None

    # -----------------------------------------------------------------------------------------------------------------
    def __init__(
            self,
            *v_args,
            _validator: TYPING.VALID_VALIDATOR = None,
            _iresult_reverse: bool = None,
            _iresult_cumulate: EnumAdj_BoolCumulate = None,
            _other_final__resolve: bool = None,
            # _other_druft: Any = None,     # DONT USE HERE! if need checking - create obj and use obj.validate(_other_druft)
            **v_kwargs,
    ) -> None:
        """
        :param _validator: dont use! define in new class as attribute
        """
        if _validator is not None:
            self.VALIDATOR = _validator

        if _iresult_reverse is not None:
            self.IRESULT_REVERSE = _iresult_reverse

        if _iresult_cumulate is not None:
            self.IRESULT_CUMULATE = _iresult_cumulate

        if _other_final__resolve is not None:
            self.OTHER_FINAL__RESOLVE = _other_final__resolve

        # super(ArgsKwargs, self).__init__(*v_args, **v_kwargs)
        self.V_ARGS = v_args
        self.V_KWARGS = v_kwargs

    def __str__(self):
        v_args = self.V_ARGS
        v_kwargs = self.V_KWARGS
        _iresult_reverse = self.IRESULT_REVERSE
        _iresult_cumulate = self.IRESULT_CUMULATE
        _other_final__resolve = self.OTHER_FINAL__RESOLVE
        return f"{self.__class__.__name__}({v_args=}/{v_kwargs=}/{_iresult_reverse=}/{_iresult_cumulate=}/{_other_final__resolve=})"

    def __repr__(self):
        return str(self)

    def VALIDATOR(self, other_final, *v_args, **v_kwargs) -> bool | NoReturn:
        return NotImplemented

    # DOUBT -----------------------------------------------------------------------------------------------------------
    def __iter__(self) -> Iterable[Any]:
        """
        NOTE
        ----
        not always correct!
        best usage for EqVariants or for any object with several args (Reqexp/...)

        # TODO: DEPRECATE??? - think no!!
        """
        yield from self.V_ARGS

    # PREPARES --------------------------------------------------------------------------------------------------------
    def _other_final__resolve(self, other_draft: Any, *other_args, **other_kwargs) -> None:
        # TODO: decide use or not callable other??? = USE! it is really need to validate callable!!!
        if self.OTHER_FINAL__RESOLVE:
            try:
                self.OTHER_FINAL = Lambda(other_draft, *other_args, **other_kwargs).resolve__raise()
                self.OTHER_RAISED = False
            except Exception as exc:
                self.OTHER_RAISED = True
                self.OTHER_FINAL = exc
        else:
            self.OTHER_FINAL = other_draft
            # self.OTHER_RAISED = False     # DONT PLACE HERE!!!

        self._other_final__push_chain()

    def _other_final__push_chain(self) -> None:
        """
        GOAL
        ---
        if object is CHAIN apply FinalValue from first chain without resolving in other chains!!!
        """
        for v_arg in self.V_ARGS:
            if isinstance(v_arg, Base_EqValid):
                v_arg.OTHER_FINAL__RESOLVE = False
                v_arg.OTHER_RAISED = self.OTHER_RAISED

    # VALIDATE-1=VALUE SINGLE -----------------------------------------------------------------------------------------
    def __eq__(self, other_draft) -> bool:
        return self.resolve(other_draft)

    def __contains__(self, item) -> bool:
        return self.resolve(item)

    # VALIDATE-2=VALUE with argsKwrgs----------------------------------------------------------------------------------
    def resolve(self, other_draft: Any, *other_args, **other_kwargs) -> bool:
        """
        GOAL
        ----
        validate smth with special logic
        """
        # OTHER_FINAL --------------------
        self._other_final__resolve(other_draft, *other_args, **other_kwargs)

        # VALIDATION --------------------
        if not self.V_ARGS:
            # 1=ARGS BLANK --------------
            validator_result = Lambda(self.VALIDATOR, self.OTHER_FINAL, **self.V_KWARGS).resolve__bool()
            if self.IRESULT_REVERSE:
                result = not validator_result
            else:
                result = validator_result
            return result

        else:
            # 2=ARGS one or MORE --------------
            for v_arg in self.V_ARGS:
                validator_result = Lambda(self.VALIDATOR, self.OTHER_FINAL, v_arg, **self.V_KWARGS).resolve__bool()
                if self.IRESULT_REVERSE:
                    result = not validator_result
                else:
                    result = validator_result

                # CUMULATE --------
                if self.IRESULT_CUMULATE == EnumAdj_BoolCumulate.ALL_TRUE:
                    if not result:
                        return False
                elif self.IRESULT_CUMULATE == EnumAdj_BoolCumulate.ANY_TRUE:
                    if result:
                        return True
                elif self.IRESULT_CUMULATE == EnumAdj_BoolCumulate.ALL_FALSE:
                    if result:
                        return False
                elif self.IRESULT_CUMULATE == EnumAdj_BoolCumulate.ANY_FALSE:
                    if not result:
                        return True

            # FINAL ------------
            if self.IRESULT_CUMULATE in [EnumAdj_BoolCumulate.ALL_TRUE, EnumAdj_BoolCumulate.ALL_FALSE]:
                return True
            else:
                return False


# =====================================================================================================================
class Base_ValidEnumValue(Base_EqValid):
    """
    GOAL
    ----
    use it as SelfComplete validator object (OTHER_DRAFT in )
    """
    # redefine!
    OTHER_DRAFT: Any | Callable

    def resolve(self, *other_args, **other_kwargs) -> bool:
        return super().resolve(self.OTHER_DRAFT, *other_args, **other_kwargs)


# =====================================================================================================================



from typing import *

from base_aux.base_values.m2_value_special import NoValue
# from base_aux.base_types.m1_type_aux import *      # RECURSION EXC
from .m1_init0_post import NestInit_Post


# =====================================================================================================================
class NestInit_Source(NestInit_Post):
    """
    GOAL
    ----
    just show that class uses init with source - one param
    means inside instance all methods will work on it!
    and all params for methods belong to methods! not about new source!

    SPECIALLY CREATED FOR
    ---------------------
    apply in AttrAux_Existed and same others

    :ivar SOURCE: use Lambda or even simple Callable to make a callableGenerate default value! like dict or other user class.
        main Idea for Callable Source is generate independent value in instances.
        it keeps callable type only in class attribute! in instance it will be resolved by calling!

    BEST USAGE
    ----------
        class ClsAux(NestInit_Source):
            SOURCE = MyClass
            SOURCE = Lambda(dict)
            SOURCE = dict
    """
    # SOURCE: dict = Lambda(dict)               # for callable
    # SOURCE: Any | Lambda = Lambda(None)       # generic final value
    SOURCE: Any = None                          # generic final value
    # SOURCE_DEF__CALL_IF_CALLABLE: bool = True   # used only for CLS.SOURCE! not for param source! really it is NOT NEED!

    @classmethod
    @property
    def SOURCE_DEF(cls) -> Any | NoReturn:
        result = cls.SOURCE

        # create independent type class
        # specially created for making indepandent __annotations__

        try:
            is_class = issubclass(cls.SOURCE, object)
            result = type(result.__name__, (result, ), {})
        except:
            pass

        # if isinstance(cls.SOURCE, Lambda):
        if callable(result):        # and cls.SOURCE_DEF__CALL_IF_CALLABLE:
            result = result()

        return result

    def __init__(self, source: Any = NoValue, *args, **kwargs) -> None | NoReturn:
        self.init_source(source)
        super().__init__(*args, **kwargs)

    def init_source(self, source: Any = NoValue) -> None | NoReturn:
        if source is not NoValue:
            self.SOURCE = source
        else:
            self.SOURCE = self.SOURCE_DEF


# =====================================================================================================================
